# -*- coding: utf-8 -*-
"""
Tools for cycle counting. Adapted and enhanced from the Yeti version.
"""

import warnings
import numpy as np
import pandas as pd
from pyyeti import locate

try:
    import numba
except ImportError:
    HAVE_NUMBA = False
    numba_bool = bool
else:
    HAVE_NUMBA = True
    numba_bool = numba.types.bool_

try:
    import pyyeti.rainflow.c_rain as rain
except ImportError:
    if not HAVE_NUMBA:
        warnings.warn(
            "Compiled C version of rainflow algorithm failed to "
            "import. Using plain Python version, which can be very "
            "slow. Recommend installing 'numba': pyYeti will use that "
            "automatically to speed up the Python version to be on "
            "par with the C version.",
            RuntimeWarning,
        )
    import pyyeti.rainflow.py_rain as rain

# FIXME: We need the str/repr formatting used in Numpy < 1.14.
try:
    np.set_printoptions(legacy="1.13")
except TypeError:
    pass


def rainflow(peaks, getoffsets=False, use_pandas=True):
    """
    Rainflow cycle counting.

    Parameters
    ----------
    peaks : 1d array_like
        Vector of alternating peaks (as returned by :func:`findap`,
        for example)
    getoffsets : bool; optional
        If True, the tuple ``(rf, os)`` is returned; otherwise, only
        `rf` is returned.
    use_pandas : bool; optional
        If True, this routine will return pandas DataFrames

    Returns
    -------
    rf : pandas DataFrame or 2d ndarray
        n x 3 matrix with the rainflow cycle count information with
        the index going from 0 to n-1 and the columns being ['amp',
        'mean', 'count']:

            - amp is the cycle amplitude (half the peak-to-peak range)
            - mean is mean of the cycle
            - count is either 0.5 or 1.0 depending on whether it's
              half or full cycle

    os : pandas DataFrame or 2d ndarray
        Only returned if `getoffsets` is True. n x 2 matrix of cycle
        offsets with index going from 0 to n-1 and the columns being
        ['start', 'stop']:

            - start is the offset into `peaks` for start of cycle
            - stop is the offset into `peaks` for end of cycle

    Notes
    -----
    This algorithm is derived from reference [#cc1]_. This routine is
    a wrapper for either the C version
    (:func:`pyyeti.rainflow.c_rain`) or the Python version
    (:func:`pyyeti.rainflow.py_rain`) of the algorithm. Both versions
    give identical results. They also run in comparable times if
    :func:`numba.jit` is available (if ``import numba`` works). If
    :func:`numba.jit` is not available, the C version is *much*
    faster.

    References
    ----------
    .. [#cc1] "Standard Practices for Cycle Counting in Fatigue
           Analysis", ASTM E 1049 - 85 (Reapproved 2005).

    Examples
    --------
    Run the example from the ASTM paper:

    >>> from pyyeti.cyclecount import rainflow
    >>> rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2])
       amp  mean  count
    0  1.5  -0.5    0.5
    1  2.0  -1.0    0.5
    2  2.0   1.0    1.0
    3  4.0   1.0    0.5
    4  4.5   0.5    0.5
    5  4.0   0.0    0.5
    6  3.0   1.0    0.5

    With offsets:

    >>> rf, os = rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2],
    ...                   getoffsets=True)
    >>> rf
       amp  mean  count
    0  1.5  -0.5    0.5
    1  2.0  -1.0    0.5
    2  2.0   1.0    1.0
    3  4.0   1.0    0.5
    4  4.5   0.5    0.5
    5  4.0   0.0    0.5
    6  3.0   1.0    0.5
    >>> os
       start  stop
    0      0     1
    1      1     2
    2      4     5
    3      2     3
    4      3     6
    5      6     7
    6      7     8

    If not using pandas:

    >>> rf, os = rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2],
    ...                   getoffsets=True, use_pandas=False)
    >>> rf
    array([[ 1.5, -0.5,  0.5],
           [ 2. , -1. ,  0.5],
           [ 2. ,  1. ,  1. ],
           [ 4. ,  1. ,  0.5],
           [ 4.5,  0.5,  0.5],
           [ 4. ,  0. ,  0.5],
           [ 3. ,  1. ,  0.5]])
    >>> os                              # doctest: +ELLIPSIS
    array([[0, 1],
           [1, 2],
           [4, 5],
           [2, 3],
           [3, 6],
           [6, 7],
           [7, 8]]...)
    """
    if getoffsets:
        rf, os = rain.rainflow(peaks, getoffsets)
        if use_pandas:
            rf = pd.DataFrame(rf, columns=["amp", "mean", "count"])
            os = pd.DataFrame(os, columns=["start", "stop"])
        return rf, os

    rf = rain.rainflow(peaks, getoffsets)
    if use_pandas:
        rf = pd.DataFrame(rf, columns=["amp", "mean", "count"])
    return rf


if not HAVE_NUMBA:

    def _findap_worker(y, tol):
        # example: [1, 2, 3, 4, 4, -2, -2, -2]

        if y.size == 1:
            return np.array([True])

        # first, find unique values (1st of series is unique)
        u = locate.find_unique(y, tol)
        # [ True,  True,  True,  True, False,  True, False, False]

        # work with unique values only:
        if np.all(u):
            yu = y
            allu = True
        else:
            yu = y[u]
            # [ 1,  2,  3,  4, -2]
            allu = False

        # find signs of slopes to locate local max/mins:
        s = np.sign(np.diff(yu))
        # [ 1,  1,  1, -1]

        # locate local max/mins:
        pv = np.ones(yu.size, bool)
        pv[1:-1] = np.abs(np.diff(s)) == 2
        if yu.size > 2:
            pv[-1] = yu[-1] != yu[-2]
        # [ True, False, False,  True,  True]

        if allu:
            return pv

        # expand to full size:
        PV = np.zeros(y.size, bool)  # non-uniques are not peaks
        PV[u] = pv
        # [ True, False, False,  True, False,  True, False, False]
        return PV

else:

    def _findap_worker(y, tol):
        if y.size == 1:
            return np.array([True])

        if y.size == 2:
            if y[1] == y[0]:
                return np.array([True, False])
            return np.array([True, True])

        stol = np.abs(tol * np.abs(np.diff(y)).max())

        PV = np.zeros(y.size, numba_bool)
        PV[0] = True
        prv = y[0]
        i = 1
        while i < y.size:
            if np.abs(y[i] - prv) > stol:
                break
            i += 1

        if i == y.size:
            return PV

        j = i
        cur = y[j]
        if cur > prv:
            mountain = True  # find mountain peak
        else:
            mountain = False  # find valley floor

        for i in range(i + 1, y.size):
            nxt = y[i]
            if np.abs(nxt - cur) > stol:
                if mountain:
                    if nxt < cur:
                        PV[j] = True
                        mountain = False
                else:
                    if nxt > cur:
                        PV[j] = True
                        mountain = True
                cur = nxt
                j = i

        if np.abs(nxt - y[-2]) > stol:
            PV[-1] = True
        else:
            PV[j] = True

        return PV


def findap(y, tol=1e-6, nan_check=True):
    """
    Find alternating local maximum and minimum points in a vector.

    Parameters
    ----------
    y : ndarray
        y-axis data vector
    tol : scalar; optional
        Tolerance value for detecting unique values; see
        :func:`pyyeti.locate.find_unique`
    nan_check : bool; optional
        Whether to check for NaNs. Set to ``False`` for a performance
        gain if you know there are no NaNs.

    Returns
    -------
    pv : ndarray
        Boolean vector for the alternating peaks in `y`.

    Notes
    -----
    `y` is flattened to one dimension before operations.

    When `y` has a series of equal points, the first of the series is
    considered the peak. The first value in `y` is always considered a
    peak. The last point is a peak if and only if it is not equal to
    the point before it.

    This routine is typically used to prepare a signal for cycle
    counting.

    NaNs are not considered to be peaks and treated as if they are not
    present when `nan_check` is True (the default).

    Examples
    --------
    >>> from pyyeti import cyclecount
    >>> import numpy as np
    >>> cyclecount.findap(np.array([1]))
    array([ True], dtype=bool)
    >>> cyclecount.findap(np.array([1, 1, 1, 1]))
    array([ True, False, False, False], dtype=bool)
    >>> tf = cyclecount.findap(np.array([1, 2, 3, 4, 4, -2, -2, 0]))
    >>> np.nonzero(tf)[0]               # doctest: +ELLIPSIS
    array([0, 3, 5, 7]...)
    >>> tf = cyclecount.findap(np.array([1, 2, 3, 4, 4, -2, -2, -2]))
    >>> np.nonzero(tf)[0]               # doctest: +ELLIPSIS
    array([0, 3, 5]...)
    >>> tf = cyclecount.findap(np.array([1, 2, 3, 4, -2]))
    >>> np.nonzero(tf)[0]               # doctest: +ELLIPSIS
    array([0, 3, 4]...)
    >>> cyclecount.findap(np.array([np.nan, 1, 2, np.nan]))
    array([False,  True,  True, False], dtype=bool)
    """
    y = y.ravel()
    if nan_check:
        nans = np.isnan(y)
        if np.any(nans):
            full_pv = np.zeros(y.size, bool)
            non_nans = ~nans
            y = y[non_nans]
            if y.size == 0:
                return full_pv

            pv = _findap_worker(y, tol)

            full_pv[non_nans] = pv

            return full_pv

    return _findap_worker(y, tol)


def getbins(bins, mx, mn, right=True, check_bounds=False):
    """
    Utility routine used by :func:`sigcount` to get bin boundaries.

    Parameters
    ----------
    bins : scalar integer or 1d array_like
        Used to define bin boundaries. If array_like, must be
        monotonically increasing. See description below.
    mx, mn : scalar
        Maximum and minimum values of data to be put in bins. `mx` and
        `mn` may be input in either order. If `bins` is a 1d
        array_like, `mx` and `mn` are ignored unless `check_bounds`
        is True; see below.
    right : bool; optional
        Indicates whether the bins include the rightmost edge or the
        leftmost edge. If right == True (the default), then the bins
        [1,2,3,4] indicate (1,2], (2,3], (3,4].
    check_bounds : bool; optional
        If True, routine will check to see if the final bin boundaries
        cover the `mn` to `mx` range. Note that the range is guaranteed to be
        covered if `bins` is a scalar.

    Returns
    -------
    bb : 1d ndarray
        The bin boundaries; ``length = bins + 1``
    out_of_bounds : bool; optional
        True if either `mx` or `mn` will fall out of range of the bins
        (according to `right`). Only returned if `check_bounds` is
        True.

    Notes
    -----
    If `mx` and `mn` are equal, they are reset::

        mx = mx + 0.5
        mn = mn - 0.5

    If `bins` is a scalar, this routine tries to mimic the behavior of
    the :func:`pandas.cut` routine::

        bb = np.linspace(mn, mx, bins+1)
        p = 0.001 * (mx - mn)
        if right:
            bb[0] -= p
        else:
            bb[-1] += p

    If `bins` is a vector, it defines the boundaries directly and is
    returned as is::

        bb = np.array(bins)

    Raises
    ------
    ValueError
        If `bins` is a vector but is not monotonically increasing

    Examples
    --------
    >>> from pyyeti import cyclecount
    >>> cyclecount.getbins(4, 12, 4)
    array([  3.992,   6.   ,   8.   ,  10.   ,  12.   ])
    >>> cyclecount.getbins(4, 12, 4, right=False)
    array([  4.   ,   6.   ,   8.   ,  10.   ,  12.008])
    >>> cyclecount.getbins([1, 2, 3, 4], 12, 4, check_bounds=True)
    (array([1, 2, 3, 4]), True)
    """
    bins = np.atleast_1d(bins)
    if mx < mn:
        mx, mn = mn, mx
    elif mx == mn:
        mx = mx + 0.5
        mn = mn - 0.5
        # raise ValueError("`mx` and `mn` must not be equal")

    if bins.size == 1:
        bins = int(bins[0])
        bb = np.linspace(mn, mx, bins + 1)
        p = 0.001 * (mx - mn)
        if right:
            bb[0] -= p
        else:
            bb[-1] += p
        out_of_bounds = False
    elif bins.ndim == 1:
        if np.any(np.diff(bins) <= 0):
            raise ValueError(
                "when `bins` is input as a vector, it "
                "must be monotonically increasing"
            )
        bb = np.array(bins)
        if check_bounds:
            if right:
                if mn <= bb[0] or mx > bb[-1]:
                    out_of_bounds = True
                else:
                    out_of_bounds = False
            else:
                if mn < bb[0] or mx >= bb[-1]:
                    out_of_bounds = True
                else:
                    out_of_bounds = False

    if check_bounds:
        return bb, out_of_bounds

    return bb


def _getlabels(form, bins):
    return [form.format(i, j) for i, j in zip(bins[:-1], bins[1:])]


# @njit(cache=True)
def _binify(cycles, bins_range, bins_mean, right, ensure_boundaries=True):
    # Many thanks to Daniel Kowollik for contributing the original
    # implementation of this routine! :)
    num_bins_mean = len(bins_mean) - 1
    num_bins_range = len(bins_range) - 1

    bin_indices_range = np.digitize(cycles[:, 0], bins_range, right=right) - 1
    bin_indices_mean = np.digitize(cycles[:, 1], bins_mean, right=right) - 1

    markov_matrix = np.zeros((num_bins_mean, num_bins_range), np.float64)
    if ensure_boundaries:
        for i in range(len(cycles)):
            bim = bin_indices_mean[i]
            bir = bin_indices_range[i]
            if (0 <= bim < num_bins_mean) and (0 <= bir < num_bins_range):
                markov_matrix[bim, bir] += cycles[i, 2]
    else:
        for i in range(len(cycles)):
            markov_matrix[bin_indices_mean[i], bin_indices_range[i]] += cycles[i, 2]

    return markov_matrix


if HAVE_NUMBA:
    _findap_worker = numba.njit(cache=True)(_findap_worker)
    _binify = numba.njit(cache=True)(_binify)

    @numba.njit(cache=True)
    def maxmin(x):
        mn = mx = x[0]
        for v in x[1:]:
            if v > mx:
                mx = v
            elif v < mn:
                mn = v
        return mx, mn

else:

    def maxmin(x):
        return x.max(), x.min()


def binify(
    rf,
    ampbins=10,
    meanbins=1,
    right=True,
    precision=3,
    retbins=False,
    use_pandas=True,
    check_bounds=True,
):
    """
    Summarize cycle count results (as from :func:`rainflow`) into
    bins.

    Parameters
    ----------
    rf : 2d array_like
        2d matrix with (at least) 3 columns: [amp, mean, count]. See
        :func:`rainflow`.
    ampbins : scalar or 1d array_like; optional
        Defines the cycle amplitude bins; see :func:`getbins` for
        more complete description on how bins are defined.

        ==============   =============================================
        `ampbins` type   Brief description (see also :func:`getbins`)
        ==============   =============================================
           scalar        Defines number of bins to use
           vector        Defines lower bin boundaries
        ==============   =============================================

    meanbins : scalar or 1d array_like; optional
        Defines the cycle mean-value bins; see `ampbins` description
        and :func:`getbins` for more complete description on how bins
        are defined.
    right : bool; optional
        Indicates whether the bins include the rightmost edge or the
        leftmost edge. If right == True (the default), then the bins
        [1,2,3,4] indicate (1,2], (2,3], (3,4].
    precision : integer; optional
        Precision to use for DataFrame labels; only used if
        `use_pandas` is True.
    retbins : bool; optional
        If True, return the `ampbins` and `meanbins` vectors.
    use_pandas : bool; optional
        If True, this routine will return `table` as a pandas
        DataFrame.
    check_bounds : bool; optional
        If True, routine will check to see if the final bin boundaries
        for both the amplitude and mean cover the range of the
        data. Note that the range is guaranteed to be covered if the
        `ampbins` and `meanbins` inputs are scalars.

    Returns
    -------
    table : pandas DataFrame or 2d ndarray
        A cycle count table (len(`meanbins`) x len(`ampbins`)). Each
        value in `table` entry is the number of cycles in the bin
        (see below).
    ampbins : 1d ndarray; optional
        Boundaries of amplitude bins used; length = # of bins + 1.
        Only returned if `retbins` is True.
    meanbins : 1d ndarray; optional
        Boundaries of mean-value bins used; length = # of bins + 1.
        Only returned if `retbins` is True.

    Notes
    -----
    Algorithm works as follows:

    For each mean value bin (i'th):
       1. `rf` is filtered down to only those rows where the mean value
          falls in the mean-bin (call this subset of `rf` `rffilt`).
       2. For each amplitude bin (j'th): count number of cycles in
          `rffilt` with amplitude that falls in the amp-bin and store in
          ``table[i, j]``

    A value falls in bin `i` if::

        bin[i] < value <= bin[i+1]        } if right is True
        bin[i] <= value < bin[i+1]        } if right is not True

    Examples
    --------
    >>> from pyyeti import cyclecount
    >>> rf = cyclecount.rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2])
    >>> rf
       amp  mean  count
    0  1.5  -0.5    0.5
    1  2.0  -1.0    0.5
    2  2.0   1.0    1.0
    3  4.0   1.0    0.5
    4  4.5   0.5    0.5
    5  4.0   0.0    0.5
    6  3.0   1.0    0.5
    >>> format = lambda x: f"{x:.1f}"
    >>> df = cyclecount.binify(rf, 3, 2)
    >>> df.map(format)                # doctest: +ELLIPSIS
    Amp             (1.497, 2.500] (2.500, 3.500] (3.500, 4.500]
    Mean...
    (-1.002, 0.000]            1.0            0.0            0.5
    (0.000, 1.000]             1.0            0.5            1.0
    >>> df = cyclecount.binify(rf, 3, 2, right=0)
    >>> df.map(format)                # doctest: +ELLIPSIS
    Amp             [1.500, 2.500) [2.500, 3.500) [3.500, 4.503)
    Mean...
    [-1.000, 0.000)            1.0            0.0            0.0
    [0.000, 1.002)             1.0            0.5            1.5

    If not using pandas:

    >>> cyclecount.binify(rf, 3, 2, use_pandas=False)
    array([[ 1. ,  0. ,  0.5],
           [ 1. ,  0.5,  1. ]])
    """
    rf = np.atleast_2d(rf)
    # rf = [amp, mean, count] columns

    ampb = getbins(ampbins, *maxmin(rf[:, 0]), right, check_bounds)
    aveb = getbins(meanbins, *maxmin(rf[:, 1]), right, check_bounds)

    if check_bounds:
        ampb, out_amp = ampb
        aveb, out_ave = aveb
        out = out_amp or out_ave
    else:
        out = False

    table = _binify(rf, ampb, aveb, right, out)

    if use_pandas:
        f = "{:." + str(precision) + "f}"
        f = f + ", " + f
        if right:
            form = "(" + f + "]"
        else:
            form = "[" + f + ")"

        index = _getlabels(form, aveb)
        columns = _getlabels(form, ampb)
        table = pd.DataFrame(table, index=index, columns=columns)
        table.columns.name = "Amp"
        table.index.name = "Mean"

    if retbins:
        return table, ampb, aveb

    return table


def sigcount(
    sig, ampbins=10, meanbins=1, right=True, precision=3, retbins=False, use_pandas=True
):
    """
    Do rainflow cycle counting on a signal.

    Parameters
    ----------
    sig : 1d array_like
        Signal (vector) to do cycle counting on.
    ampbins : scalar or 1d array_like
        Defines the cycle amplitude bins; see :func:`getbins` for
        more complete description on how bins are defined.

        ==============   =============================================
        `ampbins` type   Brief description (see also :func:`getbins`)
        ==============   =============================================
           scalar        Defines number of bins to use
           vector        Defines lower bin boundaries
        ==============   =============================================

    meanbins : scalar or 1d array_like
        Defines the cycle mean-value bins; see `ampbins` description
        and :func:`getbins` for more complete description on how bins
        are defined.
    right : bool; optional
        Indicates whether the bins include the rightmost edge or the
        leftmost edge. If right == True (the default), then the bins
        [1,2,3,4] indicate (1,2], (2,3], (3,4].
    precision : integer; optional
        Precision to use for DataFrame labels.
    retbins : bool; optional
        If True, return the `ampbins` and `meanbins` vectors.
    use_pandas : bool; optional
        If True, this routine will return `table` as a pandas
        DataFrame.

    Returns
    -------
    table : pandas DataFrame or 2d ndarray
        A cycle count table (len(`meanbins`) x len(`ampbins`)). Each
        value in `table` entry is the number of cycles in the bin
        (see below).
    ampbins : 1d ndarray
        Boundaries of amplitude bins used; length = # of bins + 1.
        Only returned if `retbins` is True.
    meanbins : 1d ndarray
        Boundaries of mean-value bins used; length = # of bins + 1.
        Only returned if `retbins` is True.

    Notes
    -----
    Steps:
      1.  Calls :func:`findap` to find all local minima and maxima:
          ``peaks = sig[findap(sig)]``
      2.  Calls :func:`rainflow` to do the cycle counting:
          ``rf = rainflow(peaks)``
      3.  Summarizes the rainflow counting by putting the counts in
          bins. For each mean value bin (i'th):

          a. `rf` is filtered down to only those rows where the mean
             value falls in the mean-bin (call this subset of `rf`
             `rffilt`).
          b. For each amplitude bin (j'th): count number of cycles in
             `rffilt` with amplitude that falls in the amp-bin and
             store in ``table[i, j]``

    A value falls in bin `i` if::

        bin[i] < value <= bin[i+1]        } if right is True
        bin[i] <= value < bin[i+1]        } if right is not True

    Examples
    --------
    >>> from pyyeti import cyclecount
    >>> import numpy as np
    >>> sig = np.arange(100)
    >>> sig[::2] *= -1   # [0, 1, -2, 3, -4, ..., 99]
    >>> # `sig` has 99 half-cycles; amplitude grows from 0.5 up to
    >>> #  98.5; mean of each is either 0.5 or -0.5
    >>> table = cyclecount.sigcount(sig, 2, 2)
    >>> table                              # doctest: +ELLIPSIS
    Amp              (0.402, 49.500]  (49.500, 98.500]
    Mean...
    (-0.501, 0.000]             12.5              12.0
    (0.000, 0.500]              12.5              12.5

    We can focus on amplitudes only (ignoring the mean value bins):

    >>> table.sum(axis=0)
    Amp
    (0.402, 49.500]     25.0
    (49.500, 98.500]    24.5
    dtype: float64

    Focus on only the mean value bins:

    >>> table.sum(axis=1)
    Mean
    (-0.501, 0.000]    24.5
    (0.000, 0.500]     25.0
    dtype: float64

    If not using pandas:

    >>> cyclecount.sigcount(sig, 2, 2, use_pandas=False)
    array([[ 12.5,  12. ],
           [ 12.5,  12.5]])
    """
    rf = rainflow(sig[findap(sig)], use_pandas=False)
    return binify(rf, ampbins, meanbins, right, precision, retbins, use_pandas)



import numpy as np
import matplotlib as mpl

mpl.interactive(0)
mpl.use("Agg")
from pyyeti import cla


def scatm(sol, nas, Vars, se):
    return Vars[se]["atm"] @ sol.a


def net_ifltm(sol, nas, Vars, se):
    return Vars[se]["net_ifltm"] @ sol.a


def net_ifatm(sol, nas, Vars, se):
    return Vars[se]["net_ifatm"] @ sol.a


def get_xyr():
    # return the xr, yr, and rr indexes for the "cglf" data recovery
    # ... see :func:`cla.DR_Defs.add`
    xr = np.array([1, 3, 6, 8])  # 'x' row(s)
    yr = xr + 1  # 'y' row(s)
    rr = np.arange(4) + 10  # rss  rows
    return xr, yr, rr


def cglf(sol, nas, Vars, se):
    resp = Vars[se]["cglf"] @ sol.a
    xr, yr, rr = get_xyr()
    resp[rr] = np.sqrt(resp[xr] ** 2 + resp[yr] ** 2)
    return resp


def cglf_psd(sol, nas, Vars, se, freq, forcepsd, drmres, case, i):
    resp = Vars[se]["cglf"] @ sol.a
    cla.PSD_consistent_rss(resp, *get_xyr(), freq, forcepsd, drmres, case, i)


if __name__ == "__main__":
    import os
    import inspect
    import re
    import matplotlib.pyplot as plt
    from pyyeti import op2, n2p, nastran, op4, cb
    from pyyeti.pp import PP

    se = 101
    uset, coords = nastran.bulk2uset("outboard.asm")
    dct = op4.read("outboard.op4")
    maa = dct["mxx"]
    kaa = dct["kxx"]
    atm = dct["mug1"]
    #    ltm = dct['mef1']
    pch = "outboard.pch"

    def getlabels(lbl, id_dof):
        return ["{} {:4d}-{:1d}".format(lbl, g, i) for g, i in id_dof]

    atm_labels = getlabels("Grid", nastran.rddtipch(pch, "tug1"))
    #    ltm_labels = getlabels('CBAR', nastran.rddtipch(pch, 'tef1'))
    iflabels = getlabels("Grid", uset[:, :2].astype(int))

    # setup CLA parameters:
    mission = "Micro Space Station"

    nb = uset.shape[0]
    nq = maa.shape[0] - nb
    bset = np.arange(nb)
    qset = np.arange(nq) + nb
    ref = [600.0, 150.0, 150.0]
    g = 9806.65
    net = cb.mk_net_drms(maa, kaa, bset, uset=uset, ref=ref, g=g)

    # run cbcheck:
    chk = cb.cbcheck(
        "outboard_cbcheck.out", maa, kaa, bset, bref=np.arange(6), uset=uset, uref=ref
    )

    # define some defaults for data recovery:
    defaults = dict(
        se=se, uf_reds=(1, 1, 1.25, 1), srsfrq=np.arange(0.1, 50.1, 0.1), srsQs=(10, 33)
    )

    drdefs = cla.DR_Def(defaults)

    @cla.DR_Def.addcat
    def _():
        name = "scatm"
        desc = "Outboard Internal Accelerations"
        units = "mm/sec^2, rad/sec^2"
        labels = atm_labels[:12]
        drms = {"atm": atm[:12]}
        prog = re.compile(" [2]-[1]")
        histpv = [i for i, s in enumerate(atm_labels) if prog.search(s)]
        srspv = histpv
        srsopts = dict(eqsine=1, ic="steady")
        drdefs.add(**locals())

    @cla.DR_Def.addcat
    def _():
        name = "cglf"
        desc = "S/C CG Load Factors"
        units = "G"
        labels = net.cglf_labels
        drms = {"cglf": net.cglfa}
        histpv = slice(1)
        drdefs.add(**locals())

    @cla.DR_Def.addcat
    def _():
        name = "net_ifatm"
        desc = "NET S/C Interface Accelerations"
        units = "g, rad/sec^2"
        labels = net.ifatm_labels[:3]
        drms = {"net_ifatm": net.ifatm[:3]}
        srsopts = dict(eqsine=1, ic="steady")
        histpv = "all"
        srspv = np.array([True, False, True])
        drdefs.add(**locals())

    @cla.DR_Def.addcat
    def _():
        name = "net_ifltm"
        desc = "NET I/F Loads"
        units = "mN, mN-mm"
        labels = net.ifltm_labels[:6]
        drms = {"net_ifltm": net.ifltma[:6]}
        drdefs.add(**locals())

    # add a 0rb version of the NET ifatm:
    drdefs.add_0rb("net_ifatm")

    # make excel summary file for checking:
    drdefs.excel_summary()

    # save data to gzipped pickle file:
    sc = dict(mission=mission, drdefs=drdefs)
    cla.save("cla_params.pgz", sc)



# Comparing results

import matplotlib as mpl

mpl.interactive(0)
mpl.use("Agg")

import numpy as np
from pyyeti import cla
from pyyeti.pp import PP

# Load both sets of results and report percent differences:
results = cla.load("results.pgz")
lvc = cla.load("contractor_results_no_srs.pgz")

results["extreme"].rptpct(lvc, names=("LSP", "Contractor"))



import numpy as np
import pandas as pd
from pyyeti import cyclecount
from pyyeti.rainflow.py_rain import rainflow
import pytest


def test_rainflow_1():
    a = np.arange(25000)
    a[::2] *= -1
    rf, os = cyclecount.rain.rainflow(a, getoffsets=True)
    rf2, os2 = rainflow(a, getoffsets=True)
    assert np.allclose(rf, rf2)
    assert np.all(os2 == os)

    # should all be half cycles, amp = .5, 1.5, ...,
    # means = [.5, -.5, .5, -.5 ...]
    amp = np.arange(len(a) - 1, dtype=float) + 0.5
    means = 0.5 * np.ones(len(a) - 1)
    means[1::2] *= -1.0
    assert np.allclose(amp, rf[:, 0])
    assert np.allclose(means, rf[:, 1])
    assert np.allclose(0.5, rf[:, 2])

    # offset starts = 0, 1, 2, ... len(a)-1
    # offset ends   = 1, 2, 3, ... len(a)
    assert np.all(np.arange(len(a) - 1) == os[:, 0])
    assert np.all(np.arange(1, len(a)) == os[:, 1])


def test_rainflow_2():
    a = np.arange(25000)
    a[::2] *= -1
    a = a[::-1]
    rf, os = cyclecount.rain.rainflow(a, getoffsets=True)
    rf2, os2 = rainflow(a, getoffsets=True)
    assert np.allclose(rf, rf2)
    assert np.all(os2 == os)

    # reverse of:
    #   should all be half cycles, amp = .5, 1.5, ...,
    #   means = [.5, -.5, .5, -.5 ...]
    amp = np.arange(len(a) - 1, dtype=float) + 0.5
    means = 0.5 * np.ones(len(a) - 1)
    means[1::2] *= -1.0
    assert np.allclose(amp[::-1], rf[:, 0])
    assert np.allclose(means[::-1], rf[:, 1])
    assert np.allclose(0.5, rf[:, 2])

    # offset starts = 0, 1, 2, ... len(a)-1
    # offset ends   = 1, 2, 3, ... len(a)
    assert np.all(np.arange(len(a) - 1) == os[:, 0])
    assert np.all(np.arange(1, len(a)) == os[:, 1])


def test_rainflow_3():
    a = np.ones(10)
    a[::2] *= -1
    rf, os = cyclecount.rain.rainflow(a, getoffsets=True)
    rf2, os2 = rainflow(a, getoffsets=True)
    assert np.allclose(rf, rf2)
    assert np.all(os2 == os)

    # should all be half cycles, amp = 1, means = 0
    assert np.allclose(1.0, rf[:, 0])
    assert np.allclose(0.0, rf[:, 1])
    assert np.allclose(0.5, rf[:, 2])

    # offset starts = 0, 1, 2, ... len(a)-1
    # offset ends   = 1, 2, 3, ... len(a)
    assert np.all(np.arange(len(a) - 1) == os[:, 0])
    assert np.all(np.arange(1, len(a)) == os[:, 1])


def test_rainflow_4():
    a = np.ones(10)
    a[::2] *= -1
    a[0] *= 2.0  # make rainflow not accept first point until the end
    rf, os = cyclecount.rain.rainflow(a, getoffsets=True)
    rf2, os2 = rainflow(a, getoffsets=True)
    assert np.allclose(rf, rf2)
    assert np.all(os2 == os)

    # should almost all full cycles, amp = 1, means = 0
    rf_shouldbe = [
        [1.0, 0.0, 1.0],
        [1.0, 0.0, 1.0],
        [1.0, 0.0, 1.0],
        [1.0, 0.0, 1.0],
        [1.5, -0.5, 0.5],
    ]  # range: -2 to +1 = 3
    os_shouldbe = [[1, 2], [3, 4], [5, 6], [7, 8], [0, 9]]

    assert np.allclose(rf_shouldbe, rf)
    assert np.allclose(os_shouldbe, os)


def test_rainflow_5():
    rf = cyclecount.rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2])
    assert isinstance(rf, pd.DataFrame)

    rf_shouldbe = np.array(
        [
            [1.5, -0.5, 0.5],
            [2.0, -1.0, 0.5],
            [2.0, 1.0, 1.0],
            [4.0, 1.0, 0.5],
            [4.5, 0.5, 0.5],
            [4.0, 0.0, 0.5],
            [3.0, 1.0, 0.5],
        ]
    )

    os_shouldbe = np.array([[0, 1], [1, 2], [4, 5], [2, 3], [3, 6], [6, 7], [7, 8]])

    assert np.allclose(rf, rf_shouldbe)

    rf = cyclecount.rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2], use_pandas=False)
    assert not isinstance(rf, pd.DataFrame)
    assert np.allclose(rf, rf_shouldbe)

    rf, os = cyclecount.rainflow([-2, 1, -3, 5, -1, 3, -4, 4, -2], getoffsets=True)
    assert isinstance(rf, pd.DataFrame)
    assert isinstance(os, pd.DataFrame)
    assert np.allclose(rf, rf_shouldbe)
    assert np.allclose(os, os_shouldbe)

    rf, os = cyclecount.rainflow(
        [-2, 1, -3, 5, -1, 3, -4, 4, -2], getoffsets=True, use_pandas=False
    )
    assert not isinstance(rf, pd.DataFrame)
    assert not isinstance(os, pd.DataFrame)
    assert np.allclose(rf, rf_shouldbe)
    assert np.allclose(os, os_shouldbe)


def test_rainflow_badinput():
    with pytest.raises(ValueError):
        cyclecount.rain.rainflow(1.0)
    with pytest.raises(ValueError):
        rainflow(1.0)
    with pytest.raises(ValueError):
        cyclecount.rain.rainflow(np.random.randn(2, 2))
    with pytest.raises(ValueError):
        rainflow(np.random.randn(2, 2))


def test_getbins_1():
    bb = cyclecount.getbins(4, 12, 4)
    assert np.allclose(bb, [3.992, 6.0, 8.0, 10.0, 12.0])
    bb = cyclecount.getbins(4, 4, 12)
    assert np.allclose(bb, [3.992, 6.0, 8.0, 10.0, 12.0])
    bb = cyclecount.getbins(4, 4, 12, right=False)
    assert np.allclose(bb, [4.0, 6.0, 8.0, 10.0, 12.008])


def test_getbins_2():
    mx = 100
    mn = 0
    bb = cyclecount.getbins(1, mx, mn)
    assert np.allclose(bb, [mn - 0.001 * (mx - mn), mx])
    bb = cyclecount.getbins(1, mx, mn, right=False)
    assert np.allclose(bb, [mn, mx + 0.001 * (mx - mn)])


def test_getbins_3():
    mx = 20.0
    mn = -5.0
    binpts = [-2.0, 4.0, 5.0, 8.0, 14.0]
    bb = cyclecount.getbins(binpts, mn, mx)
    assert np.allclose(bb, binpts)
    binpts = [0.5, 0.1, -10.0, 100]
    with pytest.raises(ValueError):
        cyclecount.getbins(binpts, mn, mx)


def test_sigcount_1():
    sig = np.arange(100)
    sig[::2] *= -1  # [0, 1, -2, 3, -4, ..., 99]
    # `sig` has 99 half-cycles; amplitude grows from 0.5 up to
    #  98.5; mean of each is either 0.5, -0.5
    table = cyclecount.sigcount(sig, 2, 2)
    #                  (0.402, 49.500]  (49.500, 98.500]
    # (-0.501, 0.000]             12.5              12.0
    # (0.000, 0.500]              12.5              12.5
    table2, ampb, aveb = cyclecount.sigcount(sig, 2, 2, retbins=True)
    assert np.all((table2 == table).values)
    assert np.all(table.index == ["(-0.501, 0.000]", "(0.000, 0.500]"])
    assert np.all(table.columns == ["(0.402, 49.500]", "(49.500, 98.500]"])
    assert np.allclose(table.values, [[12.5, 12.0], [12.5, 12.5]])
    assert np.allclose(ampb, [0.402, 49.500, 98.500])
    assert np.allclose(aveb, [-0.501, 0.000, 0.500])


def test_sigcount_2():
    sig = np.arange(100)
    sig[::2] *= -1  # [0, 1, -2, 3, -4, ..., 99]
    # `sig` has 99 half-cycles; amplitude grows from 0.5 up to
    #  98.5; mean of each is either 0.5, -0.5
    table = cyclecount.sigcount(sig, 2, 2, right=0)
    #                  [0.500, 49.500)  [49.500, 98.598)
    # [-0.500, 0.000)             12.0              12.5
    # [0.000, 0.501)              12.5              12.5
    table2, ampb, aveb = cyclecount.sigcount(sig, 2, 2, right=0, retbins=True)
    assert np.all((table2 == table).values)
    assert np.all(table.index == ["[-0.500, 0.000)", "[0.000, 0.501)"])
    assert np.all(table.columns == ["[0.500, 49.500)", "[49.500, 98.598)"])
    assert np.allclose(table.values, [[12.0, 12.5], [12.5, 12.5]])
    assert np.allclose(ampb, [0.500, 49.500, 98.598])
    assert np.allclose(aveb, [-0.500, 0.000, 0.501])


def test_sigcount_3():
    df = cyclecount.sigcount(np.array([10.0, -10.0, 10.0, -10.0]), ampbins=4)
    assert np.allclose(df.values, [[0.0, 1.5, 0.0, 0.0]])
    assert list(df.columns) == [
        "(9.499, 9.750]",
        "(9.750, 10.000]",
        "(10.000, 10.250]",
        "(10.250, 10.500]",
    ]
    assert list(df.index) == ["(-0.501, 0.500]"]

    df = cyclecount.sigcount(
        np.array([10.0, -10.0, 10.0, -10.0]), ampbins=4, right=False
    )

    assert np.allclose(df.values, [[0.0, 0.0, 1.5, 0.0]])
    assert list(df.columns) == [
        "[9.500, 9.750)",
        "[9.750, 10.000)",
        "[10.000, 10.250)",
        "[10.250, 10.501)",
    ]
    assert list(df.index) == ["[-0.500, 0.501)"]



import numpy as np
from pyyeti.nastran import op4
from scipy.io import matlab
import os
from glob import glob
import tempfile
import scipy.sparse as sp
import pytest


def _check_badname_cm(cm):
    assert (
        cm[0].message.args[0]
        == "Output4 file has matrix name: '1mat'. Changing to 'm0'."
    )

    assert (
        cm[1].message.args[0] == "Output4 file has matrix name: ''. Changing to 'm1'."
    )

    assert (
        cm[2].message.args[0] == "Output4 file has matrix name: '09'. Changing to 'm2'."
    )


def _rdop4_tst(o4):
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4")
    nocomp = [
        "cdbin",
        "rdbin",
        "csbin",
        "rsbin",
        "cd",
        "rd",
        "cs",
        "rs",
        "x100000",
        "cdbin_ascii_sparse_bigmat",
        "cdbin_ascii_sparse_nonbigmat",
        "binary_fabiola",
        "ascii_fabiola",
    ]
    nocomp = [s + ".op4" for s in nocomp]
    m = matlab.loadmat(matfile)

    badname_translate = {"m0": "rmat", "m1": "cmat", "m2": "rcmat"}

    for filename in filenames:
        # output of dir not checked, but it should work on all these files:
        if "badname" in filename:
            with pytest.warns(RuntimeWarning) as cm:
                op4.dir(filename, verbose=False)
            _check_badname_cm(cm)
        else:
            op4.dir(filename, verbose=False)

        basename = os.path.basename(filename)
        if basename in nocomp:
            continue
        if "nas_large_dim" in basename:
            continue
        if basename.startswith("big"):
            continue

        if "badname" in filename:
            with pytest.warns(RuntimeWarning) as cm:
                dct = o4.dctload(filename)
            _check_badname_cm(cm)

            with pytest.warns(RuntimeWarning) as cm:
                names, mats, forms, mtypes = o4.listload(filename)
            _check_badname_cm(cm)

            with pytest.warns(RuntimeWarning) as cm:
                names2, sizes, forms2, mtypes2 = o4.dir(filename, verbose=False)
            _check_badname_cm(cm)
        else:
            dct = o4.dctload(filename)
            names, mats, forms, mtypes = o4.listload(filename)
            names2, sizes, forms2, mtypes2 = o4.dir(filename, verbose=False)
        assert sorted(dct.keys()) == sorted(names)
        assert names == names2
        assert forms == forms2
        assert mtypes == mtypes2
        for mat, sz in zip(mats, sizes):
            assert mat.shape == sz
        for nm in dct:
            matnm = badname_translate.get(nm, nm)
            if nm[-1] == "s":
                matnm = nm[:-1]
            assert np.allclose(m[matnm], dct[nm][0])
            pos = names.index(nm)
            assert np.allclose(m[matnm], mats[pos])
            assert dct[nm][1] == forms[pos]
            assert dct[nm][2] == mtypes[pos]

        nm2 = nm = "rcmat"
        if filename.find("single") > -1:
            nm2 = "rcmats"
        if filename.find("badname") > -1:
            nm2 = "m2"
            with pytest.warns(RuntimeWarning):
                dct = o4.dctload(filename, nm2)
                name, mat, *_ = o4.listload(filename, [nm2])
        else:
            dct = o4.dctload(filename, [nm2])
            name, mat, *_ = o4.listload(filename, nm2)
        assert np.allclose(m[nm], dct[nm2][0])
        assert np.allclose(m[nm], mat[0])


def test_rdop4():
    o4 = op4.OP4()
    _rdop4_tst(o4)


def test_rdop4_zero_rowscutoff():
    o4 = op4.OP4()
    o4._rowsCutoff = 0
    _rdop4_tst(o4)


def test_rdop4_partb():
    filenames = glob("pyyeti/tests/nastran_op4_data/*other")
    file1 = filenames[0]
    filenames = filenames[1:]
    o4 = op4.OP4()
    dct = o4.dctload(file1)
    for filename in filenames:
        dct2 = o4.dctload(filename)
        assert set(dct2.keys()) == set(dct.keys())
        for nm in dct:
            for j in range(3):
                assert np.allclose(dct2[nm][j], dct[nm][j])


def test_wtop4():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    o4 = op4.OP4()
    m = matlab.loadmat(matfile)
    names = ["rmat", "cmat", "rcmat"]
    mats = []
    wtdct = {}
    for nm in names:
        mats.append(m[nm])
        wtdct[nm] = m[nm]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [
        ["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""],
        ["pyyeti/tests/nastran_op4_data/temp_le.op4", True, "<"],
        ["pyyeti/tests/nastran_op4_data/temp_be.op4", True, ">"],
    ]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        o4.write(filename, names, mats, binary=binary, endian=endian)
        names2, sizes, forms, mtypes = o4.dir(filename, verbose=False)
        assert names2 == names
        dct = o4.dctload(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])
        o4.write(filename, wtdct, binary=binary, endian=endian)
        dct = o4.dctload(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])
    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_2():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    m = matlab.loadmat(matfile)
    names = ["rmat", "cmat", "rcmat"]
    mats = []
    wtdct = {}
    for nm in names:
        mats.append(m[nm])
        wtdct[nm] = m[nm]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [
        ["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""],
        ["pyyeti/tests/nastran_op4_data/temp_le.op4", True, "<"],
        ["pyyeti/tests/nastran_op4_data/temp_be.op4", True, ">"],
    ]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        op4.write(filename, names, mats, binary=binary, endian=endian)
        names2, sizes, forms, mtypes = op4.dir(filename, verbose=False)
        assert names2 == names
        dct = op4.load(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])
        op4.save(filename, wtdct, binary=binary, endian=endian)
        dct = op4.load(filename, into="dct")
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])

        dct = op4.read(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm])
        op4.save(filename, wtdct, binary=binary, endian=endian)
        dct = op4.read(filename, into="dct")
        for nm in dct:
            assert np.allclose(m[nm], dct[nm])

    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_single():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    o4 = op4.OP4()
    m = matlab.loadmat(matfile)
    name = "rmat"
    mat = m[name]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""]]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        o4.write(
            filename, name, mat, binary=binary, endian=endian, forms=45
        )  # 45 is not actually a valid setting
        dct = o4.dctload(filename)
        for nm in dct:
            assert nm == name
            assert np.allclose(m[nm], dct[nm][0])
            assert dct[nm][1] == 45

    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_nonbigmat_binary():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    o4 = op4.OP4()
    for name in filenames:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = o4.listload(name)
        o4.write("temp.op4", data[0], data[1], sparse="nonbigmat")
        data2 = o4.listload("temp.op4")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_wtop4_bigmat_binary():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    o4 = op4.OP4()
    for name in filenames:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = o4.listload(name)
        o4.write("temp.op4", data[0], data[1], sparse="bigmat")
        data2 = o4.listload("temp.op4")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_wtop4_nonbigmat_ascii():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    o4 = op4.OP4()
    for name in filenames:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = o4.listload(name)
        o4.write("temp.op4", data[0], data[1], sparse="nonbigmat", binary=False)
        data2 = o4.listload("temp.op4")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_wtop4_bigmat_ascii():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    o4 = op4.OP4()
    for name in filenames:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = o4.listload(name)
        o4.write("temp.op4", data[0], data[1], sparse="bigmat", binary=False)
        data2 = o4.listload("temp.op4")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_wtop4_bigmat_ascii_1():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    o4 = op4.OP4()
    for name in filenames[:1]:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = o4.load(name, into="list")
        o4.write("temp.op4", data[0], data[1], sparse="bigmat", binary=False)
        data2 = o4.load("temp.op4", into="list")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_wtop4_bigmat_ascii_2():
    filenames = glob("pyyeti/tests/nastran_op4_data/*.op4") + glob(
        "pyyeti/tests/nastran_op4_data/*.op4.other"
    )
    for name in filenames[:1]:
        if "badname" in name:
            continue
        if "nas_large_dim" in name:
            continue
        data = op4.load(name, into="list")
        op4.write("temp.op4", data[0], data[1], sparse="bigmat", binary=False)
        data2 = op4.load("temp.op4", into="list")
        assert data[0] == data2[0]
        for d1, d2 in zip(data[1], data2[1]):
            assert np.all(d1 == d2)
        os.remove("temp.op4")


def test_non_float64():
    i8 = np.array([1, 2, 0, 4], np.int8)
    i16 = i8.astype(np.int16)
    i32 = i8.astype(np.int32)
    i64 = i8.astype(np.int64)
    f32 = i8.astype(np.float32)
    c32 = (f32 + 1j * f32).astype(np.complex64)
    o4 = op4.OP4()
    for mat in [i8, i16, i32, i64, f32, c32]:
        o4.write("temp.op4", dict(mat=mat))
        mat2 = o4.dctload("temp.op4", "mat")["mat"][0]
        assert np.all(mat2 == mat)
        os.remove("temp.op4")


def test_wtop4_single_save():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    o4 = op4.OP4()
    m = matlab.loadmat(matfile)
    name = "rmat"
    mat = m[name]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""]]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        o4.save(filename, name, mat, binary=binary, endian=endian)
        dct = o4.dctload(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])
    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_single_save_1():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    o4 = op4.OP4()
    m = matlab.loadmat(matfile)
    name = "rmat"
    mat = m[name]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""]]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        o4.save(filename, name, mat, binary=binary, endian=endian)
        dct = o4.load(filename, into="dct")
        for nm in dct:
            assert np.allclose(m[nm], dct[nm][0])
    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_single_2():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    m = matlab.loadmat(matfile)
    name = "rmat"
    mat = m[name]
    # write(filename, names, matrices=None,
    #       binary=True, digits=16, endian='')
    filenames = [["pyyeti/tests/nastran_op4_data/temp_ascii.op4", False, ""]]
    for item in filenames:
        filename = item[0]
        binary = item[1]
        endian = item[2]
        op4.write(filename, name, mat, binary=binary, endian=endian)
        dct = op4.read(filename)
        for nm in dct:
            assert np.allclose(m[nm], dct[nm])
    # clean up:
    for item in filenames:
        os.remove(item[0])


def test_wtop4_single_3():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    with pytest.raises(ValueError):
        op4.load(matfile, into="badstring")


def test_wtop4_single_4():
    matfile = "pyyeti/tests/nastran_op4_data/r_c_rc.mat"
    o4 = op4.OP4()
    with pytest.raises(ValueError):
        o4.load(matfile, into="badstring")


def test_forced_bigmat():
    mat = np.zeros((1000000, 1))
    o4 = op4.OP4()
    o4.save("temp.op4", dict(mat=mat), sparse="nonbigmat")
    m = o4.load("temp.op4", into="dct")
    assert np.all(mat == m["mat"][0])

    o4.save("temp.op4", dict(mat=mat), sparse="nonbigmat", binary=False)
    m = o4.load("temp.op4", into="dct")
    assert np.all(mat == m["mat"][0])
    os.remove("temp.op4")


def test_i64():
    filenames1 = ["cdbin", "rdbin", "csbin", "rsbin"]
    filenames2 = ["cd", "rd", "cs", "rs"]
    for f1, f2 in zip(filenames1, filenames2):
        dct1 = op4.load("pyyeti/tests/nastran_op4_data/" + f1 + ".op4")
        dct2 = op4.load("pyyeti/tests/nastran_op4_data/" + f2 + ".op4")
        assert set(dct1.keys()) == set(dct2.keys())
        for nm in dct1:
            for j in range(2):
                assert np.allclose(dct1[nm][j], dct2[nm][j])


def test_bad_sparse():
    matfile = "temp1.op4"
    r = 1.2
    with pytest.raises(ValueError):
        op4.save(matfile, dict(r=r), sparse="badsparsestring")
    with pytest.raises(ValueError):
        op4.save(matfile, dict(r=r), sparse="badsparsestring", binary=False)


def test_bad_dimensions():
    matfile = "temp2.op4"
    r = np.ones((2, 2, 2))
    with pytest.raises(ValueError):
        op4.save(matfile, dict(r=r))


def test_sparse_read():
    direc = "pyyeti/tests/nastran_op4_data/"
    fnames = glob(direc + "*.op4") + glob(direc + "*.other")

    for fname in fnames:
        if "badname" in fname:
            # this test works fine but it doesn't add value and it
            # triggers annoying-to-catch warnings
            continue
        if "nas_large_dim" in fname:
            continue
        m = op4.read(fname)
        if fname.endswith("cdbin_ascii_sparse_nonbigmat.op4"):
            # this has an all-zeros matrix that is written the same as
            # the dense format (so the sparse form is not used by
            # default)
            del m["c2"]
        m2 = op4.read(fname, sparse=None)
        m3 = op4.read(fname, sparse=True)
        m4 = op4.read(fname, sparse=False)
        m5 = op4.read(fname, sparse=(None, sp.coo_matrix.tocsr))
        if "bigmat" in fname:
            for k, v in m.items():
                assert sp.issparse(m2[k])
                assert sp.isspmatrix_csr(m5[k])
                assert np.all(m2[k].toarray() == v)
                assert np.all(m5[k].toarray() == v)
        else:
            for k, v in m.items():
                assert np.all(m2[k] == v)
                assert np.all(m5[k] == v)

        for k, v in m.items():
            assert sp.issparse(m3[k])
            assert np.all(m3[k].toarray() == v)
            assert np.all(m4[k] == v)


def write_read(m, binary, sparse):
    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()
    try:
        op4.write(name, m, binary=binary, sparse=sparse)
        m2 = op4.read(name, sparse=None)
    finally:
        os.remove(name)
    return m2


def test_sparse_write():
    fnames = [
        "pyyeti/tests/nastran_op4_data/cdbin.op4",
        "pyyeti/tests/nastran_op4_data/rs.op4",
        "pyyeti/tests/nastran_op4_data/r_c_rc.op4",
        "pyyeti/tests/nastran_op4_data/double_bigmat_le.op4",
        "pyyeti/tests/nastran_op4_data/double_nonbigmat_be.op4",
        "pyyeti/tests/nastran_op4_data/single_dense_be.op4",
        # the nx nastran v9 version of the following files does not
        # make NR negative ...  not really a big deal, but the tests
        # below would need to be changed. So, just using the pyYeti
        # versions directly (nastran read them fine)
        "pyyeti/tests/nastran_op4_data/big_bigmat_ascii.op4",
        "pyyeti/tests/nastran_op4_data/big_bigmat_binary.op4",
        "pyyeti/tests/nastran_op4_data/big_dense_ascii.op4",
        "pyyeti/tests/nastran_op4_data/big_dense_binary.op4",
    ]

    for fname in fnames:
        for rd_sparse in (True, False, None):
            m = op4.read(fname, sparse=rd_sparse)
            m2 = write_read(m, binary=True, sparse="dense")
            m3 = write_read(m, binary=False, sparse="dense")
            m4 = write_read(m, binary=False, sparse="nonbigmat")
            m5 = write_read(m, binary=True, sparse="nonbigmat")
            m6 = write_read(m, binary=False, sparse="bigmat")
            m7 = write_read(m, binary=True, sparse="bigmat")
            m8 = write_read(m, binary=False, sparse="auto")
            m9 = write_read(m, binary=True, sparse="auto")

            for k, v in m.items():
                if rd_sparse:
                    assert sp.issparse(v)
                elif rd_sparse is None and "bigmat" in fname:
                    if "nonbigmat" in fname and not sp.issparse(v):
                        # v could be dense if all zeros
                        assert np.all(v == 0.0)
                    else:
                        assert sp.issparse(v)
                va = v.toarray() if sp.issparse(v) else v

                assert np.all(va == m2[k])
                assert np.all(va == m3[k])

                if sp.issparse(m4[k]):
                    assert np.all(va == m4[k].toarray())
                else:
                    assert np.allclose(va, m4[k])
                    assert np.all(m4[k] == 0.0)

                if sp.issparse(m5[k]):
                    assert np.all(va == m5[k].toarray())
                else:
                    assert np.allclose(va, m5[k])
                    assert np.all(m5[k] == 0.0)

                assert sp.issparse(m6[k])
                assert np.all(va == m6[k].toarray())

                assert sp.issparse(m7[k])
                assert np.all(va == m7[k].toarray())

                if rd_sparse or (rd_sparse is None and sp.issparse(v)):
                    assert sp.issparse(m8[k])
                    assert np.all(va == m8[k].toarray())
                    assert sp.issparse(m9[k])
                    assert np.all(va == m9[k].toarray())
                else:
                    assert np.all(va == m8[k])
                    assert np.all(va == m9[k])


def test_large_sparse():
    data = [2.3, 5, -100.4]
    rows = [2, 500000, 3500000]
    cols = [3750000, 500000, 4999999]
    a = sp.csr_matrix((data, (rows, cols)), shape=(5000000, 5000000))

    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()
    try:
        op4.write(name, dict(a=a))
        a2 = op4.read(name, sparse=None)
    finally:
        os.remove(name)

    assert sp.issparse(a2["a"])
    a2 = a2["a"].tocsr()
    for i, j, v in zip(rows, cols, data):
        assert np.allclose(a2[i, j], v)


def test_large_rows_dense():
    fname = "pyyeti/tests/nastran_op4_data/x100000.op4"
    m = op4.read(fname)
    x = np.zeros((100000, 1))
    x[45678] = 1.0
    assert np.allclose(m["x"], x)


def test_premature_eof_warning():
    a = np.random.randn(10, 10)
    f = tempfile.NamedTemporaryFile(delete=False)
    fname = f.name
    f.close()

    try:
        op4.write(fname, {"a": a})
        bufr = open(fname, "rb").read()
        open(fname, "wb").write(bufr[:-6])

        with pytest.warns(RuntimeWarning, match="Premature end-of-file"):
            a2 = op4.read(fname)

    finally:
        os.remove(fname)

    assert (a2["a"] == a).all()


def test_invalid_write_name():
    a = 1.0

    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()
    with pytest.warns(
        RuntimeWarning,
        match="Matrix for output4 write has name: '4badname'. Changing to 'm0'.",
    ):
        try:
            op4.write(name, {"4badname": a})
        finally:
            os.remove(name)


def test_too_long_write_name():
    a = 1.0

    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()
    with pytest.warns(
        RuntimeWarning,
        match="Matrix for output4 write has name: 'name_is_too_long'. "
        "Truncating to 'name_is_'.",
    ):
        try:
            op4.write(name, {"name_is_too_long": a})
        finally:
            os.remove(name)


def test_empty_file_error():
    f = tempfile.NamedTemporaryFile(delete=False)
    fname = f.name
    f.close()

    try:
        with pytest.raises(RuntimeError):
            op4.load(fname)
    finally:
        os.remove(fname)


def test_fabiola_op4():
    fname1 = "pyyeti/tests/nastran_op4_data/ascii_fabiola.op4"
    fname2 = "pyyeti/tests/nastran_op4_data/binary_fabiola.op4"
    ma = op4.read(fname1)
    mb = op4.read(fname2)
    assert np.allclose(ma["maa"], mb["maa"])
    assert np.allclose(ma["ll"], mb["ll"])
    assert np.allclose(ma["dd"], mb["dd"])
    for fn in (fname1, fname2):
        names, sizes, forms, mtypes = op4.dir(fn, verbose=False)
        assert set(names) == set(ma)
        assert set(names) == set(mb)
        assert names == ["maa", "ll", "dd"]


def test_large_dimension_op4():
    for which in ("binary", "ascii"):
        fname1 = f"pyyeti/tests/nastran_op4_data/nas_large_dim_bigmat_{which}.op4"
        fname2 = f"pyyeti/tests/nastran_op4_data/nas_large_dim_nonbigmat_{which}.op4"
        fname3 = f"pyyeti/tests/nastran_op4_data/nas_large_dim_dense_{which}.op4"

        b = op4.read(fname1, sparse=True)
        n = op4.read(fname2, sparse=True)
        d = op4.read(fname3, sparse=True)

        names1, sizes1, forms1, mtypes1 = op4.dir(fname1, verbose=False)
        names2, sizes2, forms2, mtypes2 = op4.dir(fname1, verbose=False)
        names3, sizes3, forms3, mtypes3 = op4.dir(fname1, verbose=False)
        assert names1 == ["matd", "matdt", "matd22a", "matd21"]
        assert names2 == names1
        assert names3 == names1

        assert set(names1) == set(b)
        assert set(names1) == set(n)
        assert set(names1) == set(d)

        binary = True if which == "binary" else False

        fname1 = fname1.replace("nas_", "py_")
        fname2 = fname2.replace("nas_", "py_")
        fname3 = fname3.replace("nas_", "py_")

        try:
            op4.write(fname1, b, binary=binary, sparse="bigmat")
            op4.write(fname2, b, binary=binary, sparse="nonbigmat")
            op4.write(fname3, b, binary=binary, sparse="dense")

            for name in names1:
                assert (b[name] != n[name]).nnz == 0
                assert (b[name] != d[name]).nnz == 0

            pb = op4.read(fname1, sparse=True)
            pn = op4.read(fname2, sparse=True)
            pd = op4.read(fname3, sparse=True)

            for name in ("matd", "matdt", "matd21"):
                assert (pb[name] != b[name]).nnz == 0
                assert (pn[name] != n[name]).nnz == 0
                assert (pd[name] != d[name]).nnz == 0
        finally:
            os.remove(fname1)
            os.remove(fname2)
            os.remove(fname3)



import math
import numpy as np
import scipy.linalg as la
from io import StringIO
import tempfile
import os
import inspect
from pyyeti import cb, ytools, nastran
from pyyeti.nastran import op2, n2p, op4
import pytest


def test_cbreorder():
    drm = np.arange(1, 16).reshape(3, 5)
    with pytest.raises(ValueError, match="`M` must be square when `drm` is false"):
        cb.cbreorder(drm, [0, 1, 2, 3], drm=False, last=True)


def test_uset_convert():
    import numpy as np
    from pyyeti import cb
    from pyyeti.nastran import n2p

    # node 100 in basic is @ [50, 100, 150] inches
    uset = n2p.addgrid(None, 100, "b", 0, [50, 100, 150], 0)
    ref = (100, 100, 100)
    uset_conv, ref_conv = cb.uset_convert(uset, ref, "e2m")
    assert np.allclose(uset_conv.loc[(100, 1), "x":"z"], [1.27, 2.54, 3.81])
    assert np.allclose(ref_conv, [2.54, 2.54, 2.54])

    uset_back, ref_back = cb.uset_convert(uset_conv, ref_conv)
    assert np.allclose(uset_back, uset)
    assert np.allclose(ref_back, ref)

    uset_conv2, ref_conv = cb.uset_convert(uset_conv)
    assert np.allclose(uset, uset_conv2)
    assert ref_conv is None


def test_cbconvert():
    b = np.arange(6)
    drm = np.ones((1, 8))
    with pytest.raises(ValueError, match="`M` must be square when `drm` is false"):
        cb.cbconvert(drm, b, "m2e")
    b = np.arange(5)
    with pytest.raises(ValueError, match="b-set not a multiple of 6"):
        cb.cbconvert(drm, b, "m2e", drm=True)


def test_cbconvert_2():
    n = 19
    nb = 12
    mass = np.random.randn(n, n)
    b = np.arange(n - nb, n)
    m1_1 = cb.cbconvert(mass, b, "m2e")
    m1_2 = cb.cbreorder(m1_1, b)

    m2_1 = cb.cbreorder(mass, b)
    bnew = np.arange(nb)
    m2_2 = cb.cbconvert(m2_1, bnew, "m2e")
    assert np.allclose(m1_2, m2_2)
    assert np.allclose(mass[n - nb, n - nb] * 0.005710147154735817, m1_2[0, 0])


def test_cbtf():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    maa = nas["maa"][102]
    kaa = nas["kaa"][102]
    uset = nas["uset"][102]
    b = n2p.mksetpv(uset, "a", "b")
    q = ~b
    b = np.nonzero(b)[0]

    rb = n2p.rbgeom_uset(uset.iloc[b], 3)
    freq = np.arange(1.0, 80.0, 1.0)
    a = rb[:, :1]
    a2 = a.dot(np.ones((1, len(freq))))
    a3 = rb[:, 0]

    pv = np.any(maa, axis=0)
    q = q[pv]
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]
    baa1 = np.zeros_like(maa)
    baa1[q, q] = 2 * 0.05 * np.sqrt(kaa[q, q])
    baa2 = 0.1 * np.random.randn(*maa.shape)
    baa2 = baa2.dot(baa2.T)

    bb = np.ix_(b, b)

    for baa in [baa1, baa2]:
        for delq in [False, True]:
            if delq:
                m = maa[bb]
                c = baa[bb]
                k = kaa[bb]
            else:
                m = maa
                c = baa
                k = kaa

            tf = cb.cbtf(m, c, k, a, freq, b)
            tf2 = cb.cbtf(m, c, k, a2, freq, b)
            save = {}
            tf3 = cb.cbtf(m, c, k, a3, freq, b, save)
            tf4 = cb.cbtf(m, c, k, a2, freq, b, save)

            assert np.all(freq == tf.freq)
            assert np.all(freq == tf2.freq)
            assert np.all(freq == tf3.freq)
            assert np.all(freq == tf4.freq)

            assert np.allclose(tf.frc, tf2.frc)
            assert np.allclose(tf.a, tf2.a)
            assert np.allclose(tf.d, tf2.d)
            assert np.allclose(tf.v, tf2.v)

            assert np.allclose(tf.frc, tf3.frc)
            assert np.allclose(tf.a, tf3.a)
            assert np.allclose(tf.d, tf3.d)
            assert np.allclose(tf.v, tf3.v)

            assert np.allclose(tf.frc, tf4.frc)
            assert np.allclose(tf.a, tf4.a)
            assert np.allclose(tf.d, tf4.d)
            assert np.allclose(tf.v, tf4.v)

            # confirm proper solution:
            O = 2 * np.pi * freq
            velo = 1j * O * tf.d
            acce = 1j * O * velo
            f = m.dot(acce) + c.dot(velo) + k.dot(tf.d)
            assert np.allclose(acce, tf.a)
            assert np.allclose(velo, tf.v)
            assert np.allclose(f[b], tf.frc)
            if not delq:
                assert np.allclose(f[q], 0)

    with pytest.raises(ValueError, match="`a` is not compatibly sized with `freq`"):
        cb.cbtf(maa, baa1, kaa, a2[:, :3], freq, b)

    with pytest.raises(
        ValueError, match="number of rows in `a` not compatible with `bset`"
    ):
        cb.cbtf(maa, baa1, kaa, a2[:3, :], freq, b)


def test_cbreorder_m():
    m = np.dot(np.arange(1, 9).reshape(-1, 1), np.arange(2, 10).reshape(1, -1))
    # array([[ 2,  3,  4,  5,  6,  7,  8,  9],
    #        [ 4,  6,  8, 10, 12, 14, 16, 18],
    #        [ 6,  9, 12, 15, 18, 21, 24, 27],
    #        [ 8, 12, 16, 20, 24, 28, 32, 36],
    #        [10, 15, 20, 25, 30, 35, 40, 45],
    #        [12, 18, 24, 30, 36, 42, 48, 54],
    #        [14, 21, 28, 35, 42, 49, 56, 63],
    #        [16, 24, 32, 40, 48, 56, 64, 72]])

    mnew = cb.cbreorder(m, np.arange(7, 1, -1))
    sbe = np.array(
        [
            [72, 64, 56, 48, 40, 32, 16, 24],
            [63, 56, 49, 42, 35, 28, 14, 21],
            [54, 48, 42, 36, 30, 24, 12, 18],
            [45, 40, 35, 30, 25, 20, 10, 15],
            [36, 32, 28, 24, 20, 16, 8, 12],
            [27, 24, 21, 18, 15, 12, 6, 9],
            [9, 8, 7, 6, 5, 4, 2, 3],
            [18, 16, 14, 12, 10, 8, 4, 6],
        ]
    )
    assert np.all(mnew == sbe)

    with pytest.warns(RuntimeWarning, match="b-set not a multiple of 6"):
        mnew = cb.cbreorder(m, np.arange(7, -1, -1))

    sbe = np.array(
        [
            [72, 64, 56, 48, 40, 32, 24, 16],
            [63, 56, 49, 42, 35, 28, 21, 14],
            [54, 48, 42, 36, 30, 24, 18, 12],
            [45, 40, 35, 30, 25, 20, 15, 10],
            [36, 32, 28, 24, 20, 16, 12, 8],
            [27, 24, 21, 18, 15, 12, 9, 6],
            [18, 16, 14, 12, 10, 8, 6, 4],
            [9, 8, 7, 6, 5, 4, 3, 2],
        ]
    )
    assert np.all(mnew == sbe)


def test_cbreorder_drm():
    drm = np.dot(np.arange(1, 7).reshape(-1, 1), np.arange(2, 10).reshape(1, -1))
    # array([[ 2,  3,  4,  5,  6,  7,  8,  9],
    #        [ 4,  6,  8, 10, 12, 14, 16, 18],
    #        [ 6,  9, 12, 15, 18, 21, 24, 27],
    #        [ 8, 12, 16, 20, 24, 28, 32, 36],
    #        [10, 15, 20, 25, 30, 35, 40, 45],
    #        [12, 18, 24, 30, 36, 42, 48, 54]])

    dnew = cb.cbreorder(drm, np.arange(7, 1, -1), drm=True)
    sbe = np.array(
        [
            [9, 8, 7, 6, 5, 4, 2, 3],
            [18, 16, 14, 12, 10, 8, 4, 6],
            [27, 24, 21, 18, 15, 12, 6, 9],
            [36, 32, 28, 24, 20, 16, 8, 12],
            [45, 40, 35, 30, 25, 20, 10, 15],
            [54, 48, 42, 36, 30, 24, 12, 18],
        ]
    )
    assert np.all(dnew == sbe)

    with pytest.raises(ValueError, match="`M` must be square when `drm` is false"):
        cb.cbreorder(drm, np.arange(7, 1, -1))

    with pytest.warns(RuntimeWarning, match="b-set not a multiple of 6"):
        dnew = cb.cbreorder(drm, np.arange(7, -1, -1), drm=True)
    assert np.all(drm[:, ::-1] == dnew)

    drm = np.arange(1, 16).reshape(3, 5)
    # array([[ 1,  2,  3,  4,  5],
    #        [ 6,  7,  8,  9, 10],
    #        [11, 12, 13, 14, 15]])
    with pytest.warns(RuntimeWarning, match="b-set not a multiple of 6"):
        dnew = cb.cbreorder(drm, [0, 1, 2, 3], drm=True, last=True)

    sbe = np.array([[5, 1, 2, 3, 4], [10, 6, 7, 8, 9], [15, 11, 12, 13, 14]])
    assert np.all(dnew == sbe)


def test_cgmass():
    mass = np.array(
        [
            [3, 0, 0, 0, 0, 0],
            [0, 3, 0, 0, 0, 120],
            [0, 0, 3, 0, -120, 0],
            [0, 0, 0, 1020, -60, 22],
            [0, 0, -120, -60, 7808, 23],
            [0, 120, 0, 22, 23, 7800],
        ]
    )
    mcg1, dcg1 = cb.cgmass(mass)
    sbe = np.array(
        [
            [3.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 3.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 3.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1020.0, -60.0, 22.0],
            [0.0, 0.0, 0.0, -60.0, 3008.0, 23.0],
            [0.0, 0.0, 0.0, 22.0, 23.0, 3000.0],
        ]
    )
    assert np.allclose(mcg1, sbe)
    assert np.allclose(dcg1, [40.0, 0.0, 0.0])

    mcg, dcg, gyr, pgyr, I, pI = cb.cgmass(mass, all6=True)
    assert np.all(mcg == mcg1)
    assert np.all(dcg == dcg1)
    assert np.all(abs(gyr - [18.4391, 31.6649, 31.6228]) < 1e-3)
    assert np.all(abs(pgyr - [18.4204, 31.5288, 31.7693]) < 1e-3)

    sbe = np.array([[1020.0, -60.0, 22.0], [-60.0, 3008.0, 23.0], [22.0, 23.0, 3000.0]])
    assert np.allclose(I, sbe)
    sbe = np.array(
        [[1017.9312, 0.0, 0.0], [0.0, 2982.2045, 0.0], [0.0, 0.0, 3027.8643]]
    )
    assert np.all(abs(pI - sbe) < 1e-3)

    mass[1, 0] = 4.0
    with pytest.raises(ValueError, match="mass matrix is not symmetric"):
        cb.cgmass(mass)


def gettable(lines, j, col=0, label=None, skip=0):
    table = []
    if label:
        while j < len(lines) and -1 == lines[j].find(label):
            j += 1
    j += skip
    for line in lines[j:]:
        line = line.rstrip()
        if len(line) == 0:
            break
        line = line.replace(",", " ")
        table.append([float(item) for item in line[col:].split()])
    return np.array(table), j + len(table)


def comptable(s1, s2, j1, j2, col=0, label=None, skip=0, sort2=0, truncate_rows=False):
    table1, nj1 = gettable(s1, j1[0], col, label, skip)
    table2, nj2 = gettable(s2, j2[0], col, label, skip)
    if sort2:
        table2 = np.sort(table2, axis=1)
    j1[0] = nj1
    j2[0] = nj2
    if truncate_rows:
        n = min(table1.shape[0], table2.shape[0])
        table1 = table1[:n]
        table2 = table2[:n]
    return np.allclose(table1, table2, atol=1e-1)


def compare_cbcheck_output(s, sy):
    j = [15]
    jy = [15]
    assert comptable(s, sy, j, jy, label=" ID ", skip=2)
    assert comptable(s, sy, j, jy, label="----------", skip=1)
    assert comptable(s, sy, j, jy, label="----------", skip=1)
    assert comptable(s, sy, j, jy, label="6x6 ", skip=2)
    assert comptable(s, sy, j, jy, label="6x6 ", skip=2)
    assert comptable(s, sy, j, jy, label="6x6 ", skip=2)

    assert comptable(s, sy, j, jy, label="Distance to CG", skip=4, col=23)
    assert comptable(s, sy, j, jy, label=" gyration", skip=4, col=23)
    assert comptable(s, sy, j, jy, label=" gyration", skip=4, col=23, sort2=True)

    assert comptable(s, sy, j, jy, label="Inertia ", skip=2)
    assert comptable(s, sy, j, jy, label="Principal ", skip=2)

    assert comptable(s, sy, j, jy, label="Inertia ", skip=2)
    assert comptable(s, sy, j, jy, label="Principal ", skip=2)

    assert comptable(s, sy, j, jy, label="Inertia ", skip=2)
    assert comptable(s, sy, j, jy, label="Principal ", skip=2)

    assert comptable(s, sy, j, jy, label="-----------", skip=1, col=8)
    assert comptable(s, sy, j, jy, label="Summation", skip=2, col=8)

    assert comptable(s, sy, j, jy, label="-----------", skip=1, col=8)
    assert comptable(s, sy, j, jy, label="Summation", skip=2, col=8)

    assert comptable(s, sy, j, jy, label="-----------", skip=1, col=8)
    assert comptable(s, sy, j, jy, label="Summation", skip=2, col=8)

    assert comptable(s, sy, j, jy, label="Mode", skip=2, col=8, truncate_rows=True)

    mef1 = gettable(s, j[0], label="Mode No.", skip=2)[0]
    mef2 = gettable(sy, jy[0], label="Mode No.", skip=2)[0]
    # trim mef2 down:
    pv = np.any(mef2[:, 2:] >= 2.0, axis=1)
    mef2 = mef2[pv]
    assert np.allclose(mef1, mef2)

    assert comptable(s, sy, j, jy, label="Total Eff", skip=0, col=21)


def test_cbcheck_indeterminate():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")
    q = ~b
    b = np.nonzero(b)[0]

    # write and read a file:
    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()
    out = cb.cbcheck(name, maa, kaa, b, b[:6], uset=uset, em_filt=2)
    with open(name) as f:
        sfile = f.read()
    os.remove(name)
    assert (out.m == maa).all()
    assert (out.k == kaa).all()
    assert out.uset.equals(usetb)

    rbg = n2p.rbgeom_uset(out.uset)
    assert np.allclose(rbg, out.rbg)
    rbg_s = np.vstack((la.solve(rbg[:6].T, rbg.T).T, np.zeros((q[q].size, 6))))
    assert abs(out.rbs - rbg_s).max() < 1e-5
    assert abs(out.rbe - rbg_s).max() < 1e-5

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, b[:6], usetb, em_filt=2)
        s = f.getvalue()

    assert sfile[:15] == s[:15]
    s = s.splitlines()
    sfile = sfile.splitlines()
    compare_cbcheck_output(s, sfile)

    with open("pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_101.out") as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Mass matrix is positive-definite."
    assert s[2] == "Stiffness matrix is symmetric."

    compare_cbcheck_output(s, sy)

    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, b[:6], usetb, em_filt=2, conv="e2m")
        out2 = cb.cbcheck(f, out.m, out.k, b, b[:6], out.uset, em_filt=2, conv="m2e")
        assert np.allclose(out2.uset, usetb)
        assert np.allclose(maa, out2.m)
        assert np.allclose(kaa, out2.k)

    # check for error catches:
    with StringIO() as f:
        with pytest.raises(ValueError, match="number of rows in `uset`"):
            cb.cbcheck(f, maa, kaa, b, b[:6], usetb.iloc[:-6], em_filt=2)


def test_cbcheck_determinate():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")

    q = ~b
    b = np.nonzero(b)[0]
    q = np.nonzero(q)[0]

    center = np.mean(usetb.iloc[::6, 1:], axis=0)
    rb = n2p.rbgeom_uset(usetb, center.values)

    # transform to single pt on centerline:
    # [b, q]_old = T*[b, q]_new
    #            = [[rb, 0], [0, I]] * [b, q]_new
    T = np.zeros((len(b) + len(q), 6 + len(q)))
    T[: len(b), :6] = rb
    T[len(b) :, 6:] = np.eye(len(q))

    kaa = T.T @ kaa @ T
    maa = T.T @ maa @ T
    b = np.arange(6)

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, b[:6], em_filt=2)
        s = f.getvalue()

    s = s.splitlines()
    with open(
        "pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_101_single.out"
    ) as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Mass matrix is positive-definite."
    assert s[2].startswith("Warning: stiffness matrix is not symmetric:")

    j = [10]
    jy = [10]
    assert comptable(s, sy, j, jy, label="KBB =", skip=1)
    compare_cbcheck_output(s, sy)

    # check with no em filter:
    with StringIO() as f:
        out2 = cb.cbcheck(f, maa, kaa, b, b[:6])
        s2 = f.getvalue()
    s2 = s2.splitlines()

    s_unique = [i for i in s if i not in s2]
    # ['Printing only the modes with at least 2.0% effective mass.',
    #  'The sum includes all modes.']

    s2_unique = [i for i in s2 if i not in s]
    # ['     5          7.025        0.00    0.00    0.00    0.00    0.88    0.00',
    #  '     6          7.025        0.00    0.00    0.00    0.00    0.00    0.00',
    #  '     7         10.913        0.00    0.00    0.00    0.00    0.00    1.52',
    #  '    11         25.135        0.00    0.00    0.00    0.00    0.00    0.42',
    #  '    12         25.140        1.03    0.00    0.00    0.00    0.00    0.00',
    #  '    13         42.173        0.00    0.00    0.00    0.51    0.00    0.00',
    #  '    14         42.193        0.00    0.00    1.04    0.00    1.02    0.00',
    #  '    16         46.895        0.00    0.00    0.00    0.00    0.00    0.00',
    #  '    17         69.173        0.00    0.13    0.00    0.00    0.00    0.99']

    assert len(s2) > len(s)
    assert "Printing only the modes with at least 2.0% effective mass." in s_unique
    assert "The sum includes all modes." in s_unique
    assert len(s2_unique) == len(s_unique) + 7

    # make k perfect:
    kaa[b] = 0.0
    kaa[:, b] = 0.0

    # write to a string:
    with StringIO() as f:
        out2 = cb.cbcheck(f, maa, kaa, b, b[:6], em_filt=2)

    for key, val1 in out.__dict__.items():
        val2 = getattr(out2, key)
        if key == "k":
            assert np.allclose(val1[6:, 6:], val2[6:, 6:], atol=1e-6)
        else:
            assert np.allclose(val1, val2, atol=1e-6)


def test_cbcheck_unit_convert():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")

    q = ~b
    b = np.nonzero(b)[0]
    q = np.nonzero(q)[0]

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(
            f,
            maa,
            kaa,
            b,
            b[:6],
            uset=usetb,
            em_filt=2,
            conv=[1 / 25.4, 0.005710147154735817],
            uref=[600, 150, 150],
            rb_norm=False,
        )
        s = f.getvalue()

    s = s.splitlines()
    with open(
        "pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_101_unitconv.out"
    ) as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Mass matrix is positive-definite."
    assert s[2] == "Stiffness matrix is symmetric."
    compare_cbcheck_output(s, sy)


def test_cbcheck_reorder():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = np.nonzero(n2p.mksetpv(uset, "a", "b"))[0]

    maa = cb.cbreorder(maa, b, last=True)
    kaa = cb.cbreorder(kaa, b, last=True)
    b += maa.shape[0] - len(b)

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, b[:6], usetb, em_filt=2)
        s = f.getvalue()
    s = s.splitlines()
    with open("pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_101.out") as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Mass matrix is positive-definite."
    assert s[2] == "Stiffness matrix is symmetric."
    compare_cbcheck_output(s, sy)


def test_cbcheck_indeterminate_rb_norm():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")
    q = ~b
    b = np.nonzero(b)[0]

    bref = n2p.mkdofpv(usetb, "b", [[3, 12356], [19, 3]])[0]

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, bref, usetb, em_filt=2)
        s = f.getvalue()
    s = s.splitlines()

    with open(
        "pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_101_rbnorm.out"
    ) as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Mass matrix is positive-definite."
    assert s[2] == "Stiffness matrix is symmetric."

    compare_cbcheck_output(s, sy)


def test_cbcheck_indeterminate_rb_norm2():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 102
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")
    q = ~b
    b = np.nonzero(b)[0]

    bref = n2p.mkdofpv(usetb, "b", [[3, 12356], [19, 3]])[0]

    # write to a string:
    with StringIO() as f:
        out = cb.cbcheck(f, maa, kaa, b, bref, usetb, em_filt=2, rb_norm=True)
        s = f.getvalue()
    s = s.splitlines()

    with open(
        "pyyeti/tests/nas2cam_csuper/yeti_outputs/cbcheck_yeti_102_rbnorm.out"
    ) as f:
        sy = f.read().splitlines()

    assert s[0] == "Mass matrix is symmetric."
    assert s[1] == "Warning: mass matrix is not positive-definite."
    msg = "Warning: stiffness matrix is not symmetric:"
    assert s[2].startswith(msg) or s[4].startswith(msg)

    compare_cbcheck_output(s, sy)


def test_cbcoordchk():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    kaa = kaa[pv]

    uset = nas["uset"][se]
    b = n2p.mksetpv(uset, "a", "b")
    b = np.nonzero(b)[0]

    chk0 = cb.cbcoordchk(kaa, b, b[-6:], verbose=False)
    rbmodes0 = chk0.rbmodes[b]

    # maa = cb.cbreorder(maa, b, last=True)
    kaa = cb.cbreorder(kaa, b, last=True)
    b += kaa.shape[0] - len(b)
    bref = b[-6:]

    chk1 = cb.cbcoordchk(kaa, b, bref, verbose=False)
    rbmodes1 = chk1.rbmodes[b]

    assert np.allclose(chk1.coords, chk0.coords)
    assert np.allclose(rbmodes1, rbmodes0)
    assert np.allclose(chk1.maxerr, chk0.maxerr)
    assert chk0.refpoint_chk == chk1.refpoint_chk == "pass"
    assert abs(chk0.maxerr).max() < 1e-5

    # a case where the refpoint_chk should be 'fail':
    with pytest.warns(la.LinAlgWarning, match=r"Ill\-conditioned matrix"):
        chk2 = cb.cbcoordchk(kaa, b, [25, 26, 27, 31, 32, 33], verbose=False)
    assert chk2.refpoint_chk == "fail"


def test_rbmultchk():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    maa = maa[pv]
    kaa = kaa[pv]

    uset = nas["uset"][se]
    bset = n2p.mksetpv(uset, "p", "b")
    usetb = nas["uset"][se].iloc[bset]
    b = n2p.mksetpv(uset, "a", "b")
    q = ~b
    b = np.nonzero(b)[0]

    grids = [[11, 123456], [45, 123456], [60, 123456]]
    drm101, dof101 = n2p.formtran(nas, 101, grids)

    # write and read a file:
    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()

    rb = n2p.rbgeom_uset(usetb)
    cb.rbmultchk(name, drm101, "DRM101", rb)
    with open(name) as f:
        sfile = f.read()
    os.remove(name)

    # test rbscale and unit scale:
    with StringIO() as f:
        cb.rbmultchk(f, 0.00259 * drm101, "DRM101", 100 * rb)
        s = f.getvalue()

    pos = s.find(" which is: ")
    pos2 = s[pos:].find("\n")
    assert math.isclose(float(s[pos + 10 : pos + pos2]), 100)
    s = s.splitlines()
    table, nj1 = gettable(s, 15, 0, "Absolute Maximums", 3)
    sbe = np.array(
        [
            [600, 300, 300, 0.00259],
            [600, 300, 300, 0.00259],
            [600, 300, 300, 0.00259],
            [150, -930, 150, 0.00259],
            [600, 300, 300, 0.00259],
            [150, -930, 150, 0.00259],
        ]
    )

    assert np.allclose(table[:, 1:5], sbe)

    # write to a string:
    with StringIO() as f:
        cb.rbmultchk(f, drm101, "DRM101", rb)
        s = f.getvalue()
    assert sfile == s

    # add q-set rows to rb:
    nq = np.count_nonzero(q)
    rb2 = np.vstack((rb, np.zeros((nq, 6))))
    with StringIO() as f:
        cb.rbmultchk(f, drm101, "DRM101", rb2)
        s2 = f.getvalue()
    assert s2 == s

    # check results when b-set are last:
    drm101_last = np.hstack((drm101[:, q], drm101[:, b]))
    with StringIO() as f:
        cb.rbmultchk(f, drm101_last, "DRM101", rb, bset="last")
        s2 = f.getvalue()
    assert s2 == s

    # check results when b-set are last ... using pv:
    with StringIO() as f:
        bsetpv = np.zeros((len(b) + nq), bool)
        bsetpv[-len(b) :] = True
        cb.rbmultchk(f, drm101_last, "DRM101", rb, bset=bsetpv)
        s2 = f.getvalue()
    assert s2 == s

    with StringIO() as f:
        with pytest.raises(ValueError, match="invalid `bset` string"):
            cb.rbmultchk(f, drm101, "asdf", rb, bset="bad string")

    # trim q-set columns out of drm:
    labels = [str(i[0]) + "  " + str(i[1]) for i in dof101]
    with StringIO() as f:
        cb.rbmultchk(
            f,
            drm101[:, b],
            "DRM101",
            rb,
            drm2=drm101[:, b],
            prtnullrows=True,
            labels=labels,
        )
        s2 = f.getvalue()

    # row 16 is now all zeros ... not comparable
    drm2 = drm101.copy()
    drm2[15] = 0
    with StringIO() as f:
        cb.rbmultchk(f, drm2, "DRM101", rb, drm2=drm2, prtnullrows=True, labels=labels)
        s3 = f.getvalue()
    assert s2 == s3

    s = s.splitlines()
    with open("pyyeti/tests/nas2cam_csuper/yeti_outputs/rbmultchk_yeti_101.out") as f:
        sy = f.read().splitlines()

    j = [2]
    jy = [2]
    assert comptable(s, sy, j, jy, label="Extreme ", skip=3, col=11)
    assert comptable(s, sy, j, jy, label=" Row ", skip=2, col=68)
    assert comptable(s, sy, j, jy, label=" Row ", skip=2)
    assert comptable(s, sy, j, jy, label=" Row ", skip=2)


def test_rbmultchk2():
    # write to a string:
    with StringIO() as f:
        cb.rbmultchk(
            f, np.zeros((15, 6)), "DRM", np.eye(6), drm2=np.random.randn(15, 6)
        )
        s = f.getvalue()
    assert s.find(" -- no coordinates detected --") > -1
    assert s.find("All rows in DRM are NULL") > -1
    assert s.find("There are no NULL rows in DRM2.") > -1

    with StringIO() as f:
        cb.rbmultchk(
            f, np.zeros((15, 6)), "DRM", np.eye(6), drm2=np.random.randn(14, 6)
        )
        s = f.getvalue()
    assert s.find("Error: incorrectly sized DRM2") > -1

    drm = np.random.randn(14, 6)
    drm2 = np.random.randn(14, 6)
    drm[::3] = 0.0
    drm2[1::3] = 0.0
    with StringIO() as f:
        cb.rbmultchk(f, drm, "DRM", np.eye(6), drm2=drm2)
        s = f.getvalue()
    assert s.find("different set of NULL rows") > -1

    drm = np.random.randn(14, 6)
    drm2 = np.random.randn(14, 6)
    drm[::3] = 0.0
    drm2[2::3] = 0.0
    with StringIO() as f:
        cb.rbmultchk(f, drm, "DRM", np.eye(6), drm2=drm2)
        s = f.getvalue()
    assert s.find("different set of NULL rows") > -1

    with StringIO() as f:
        with pytest.raises(ValueError, match="`rb` does not have 6 columns"):
            cb.rbmultchk(f, drm, "drm", 1)

    with StringIO() as f:
        with pytest.raises(ValueError, match="failed to get scale of rb modes"):
            cb.rbmultchk(f, drm, "drm", np.zeros((6, 6)))


def test_rbdispchk():
    coords = np.array([[0, 0, 0], [1, 2, 3], [4, -5, 25]])
    rb = n2p.rbgeom(coords)
    xyz_pv = ytools.mkpattvec([0, 1, 2], 3 * 6, 6).ravel()
    rbtrimmed = rb[xyz_pv]
    # array([[  1.,   0.,   0.,   0.,   0.,   0.],
    #        [  0.,   1.,   0.,   0.,   0.,   0.],
    #        [  0.,   0.,   1.,   0.,   0.,   0.],
    #        [  1.,   0.,   0.,   0.,   3.,  -2.],
    #        [  0.,   1.,   0.,  -3.,   0.,   1.],
    #        [  0.,   0.,   1.,   2.,  -1.,   0.],
    #        [  1.,   0.,   0.,   0.,  25.,   5.],
    #        [  0.,   1.,   0., -25.,   0.,   4.],
    #        [  0.,   0.,   1.,  -5.,  -4.,   0.]])

    with StringIO() as f:
        coords_out, maxerr = cb.rbdispchk(f, rbtrimmed)
        s = f.getvalue()

    sbe = [
        "",
        "Coordinates Determined from Rigid-Body Displacements:",
        "         Node      X         Y         Z         Error",
        "        ------  --------  --------  --------   ----------",
        "             1      0.00      0.00      0.00   0.0000e+00",
        "             2      1.00      2.00      3.00   0.0000e+00",
        "             3      4.00     -5.00     25.00   0.0000e+00",
        "",
        "Maximum absolute coordinate location error:    0 units",
        "",
        "",
    ]

    assert s == "\n".join(sbe)

    with StringIO() as f:
        coords_out, maxerr = cb.rbdispchk(f, rbtrimmed, grids=[100, 200, 300])
        s = f.getvalue()

    sbe = [
        "",
        "Coordinates Determined from Rigid-Body Displacements:",
        "         Node      ID        X         Y         Z         Error",
        "        ------  --------  --------  --------  --------   ----------",
        "             1       100      0.00      0.00      0.00   0.0000e+00",
        "             2       200      1.00      2.00      3.00   0.0000e+00",
        "             3       300      4.00     -5.00     25.00   0.0000e+00",
        "",
        "Maximum absolute coordinate location error:    0 units",
        "",
        "",
    ]

    assert s == "\n".join(sbe)

    # add a little error:
    # array([[  1.,   0.,   0.,   0.,   0.,   0.],
    #        [  0.,   1.,   0.,   0.,   0.,   0.],
    #        [  0.,   0.,   1.,   0.,   0.,   0.],
    #        [  1.,   0.,   0.,   0.,   3.,  -2.],
    #        [  0.,   1.,   0.,  -3.,   0.,   1.],
    #        [  0.,   0.,   1.,   2.,  -1.,   0.],
    #        [  1.,   0.,   0.,   0.,  25.,   5.],
    #        [  0.,   1.,   0., -25.,   0.,   4.],
    #        [  0.,   0.,   1.,  -5.,  -4.,   0.]])
    rbtrimmed[7, 3] = -25.006
    with StringIO() as f:
        coords_out, maxerr = cb.rbdispchk(f, rbtrimmed)
        s = f.getvalue()

    sbe = [
        "Warning: deviation from standard pattern, node #3 "
        "starting at row 7. \tMax deviation = 0.006 units.",
        "  Rigid-Body Rotations:",
        "    0.0000    25.0000     5.0000",
        "  -25.0060     0.0000     4.0000",
        "   -5.0000    -4.0000     0.0000",
        "",
        "",
        "Coordinates Determined from Rigid-Body Displacements:",
        "         Node      X         Y         Z         Error",
        "        ------  --------  --------  --------   ----------",
        "             1      0.00      0.00      0.00   0.0000e+00",
        "             2      1.00      2.00      3.00   0.0000e+00",
        "             3      4.00     -5.00     25.00   6.0000e-03",
        "",
        "Maximum absolute coordinate location error:  0.006 units",
        "",
        "",
    ]

    assert s == "\n".join(sbe)

    with StringIO() as f:
        coords_out, maxerr = cb.rbdispchk(f, rbtrimmed, grids=[100, 200, 300])
        s = f.getvalue()

    sbe = [
        "Warning: deviation from standard pattern, node ID = 300 "
        "starting at row 7. Max deviation = 0.006 units.",
        "  Rigid-Body Rotations:",
        "    0.0000    25.0000     5.0000",
        "  -25.0060     0.0000     4.0000",
        "   -5.0000    -4.0000     0.0000",
        "",
        "",
        "Coordinates Determined from Rigid-Body Displacements:",
        "         Node      ID        X         Y         Z         Error",
        "        ------  --------  --------  --------  --------   ----------",
        "             1       100      0.00      0.00      0.00   0.0000e+00",
        "             2       200      1.00      2.00      3.00   0.0000e+00",
        "             3       300      4.00     -5.00     25.00   6.0000e-03",
        "",
        "Maximum absolute coordinate location error:  0.006 units",
        "",
        "",
    ]

    assert s == "\n".join(sbe)

    f = tempfile.NamedTemporaryFile(delete=False)
    name = f.name
    f.close()

    coords_out, maxerr = cb.rbdispchk(name, rbtrimmed, grids=[100, 200, 300])
    with open(name) as f:
        s2 = f.read()
    os.remove(name)
    assert s == s2

    with StringIO() as f:
        with pytest.raises(ValueError, match="`rbdisp` does not have 6 columns"):
            cb.rbdispchk(f, rbtrimmed[:, :4])
        with pytest.raises(ValueError, match="number of rows in `rbdisp`"):
            cb.rbdispchk(f, rbtrimmed[:5, :])


def test_cbcoordchk2():
    k = np.random.randn(14, 14)
    k = k.dot(k.T)
    b = np.arange(4)
    with pytest.raises(ValueError, match="b-set not a multiple of 6"):
        cb.cbcoordchk(k, b, b)

    b2 = np.arange(6)
    with pytest.raises(ValueError, match="reference point must have length of 6"):
        cb.cbcoordchk(k, b2, b)


def test_cbcoordchk3():
    nas = op2.rdnas2cam("pyyeti/tests/nas2cam_csuper/nas2cam")
    se = 101
    maa = nas["maa"][se]
    kaa = nas["kaa"][se]
    pv = np.any(maa, axis=0)
    pv = np.ix_(pv, pv)
    kaa = kaa[pv]

    uset = nas["uset"][se]
    b = n2p.mksetpv(uset, "a", "b")
    b = np.nonzero(b)[0]

    chk0 = cb.cbcoordchk(kaa, b, b[-6:], verbose=False)
    rbmodes0 = chk0.rbmodes[b]

    # maa = cb.cbreorder(maa, b, last=True)
    kaa = cb.cbreorder(kaa, b, last=True)
    b += kaa.shape[0] - len(b)
    bref = b[-6:]

    chk1 = cb.cbcoordchk(kaa, b, bref, verbose=False)
    rbmodes1 = chk1.rbmodes[b]

    assert np.allclose(chk1.coords, chk0.coords)
    assert np.allclose(rbmodes1, rbmodes0)
    assert np.allclose(chk1.maxerr, chk0.maxerr)
    assert chk0.refpoint_chk == chk1.refpoint_chk == "pass"
    assert abs(chk0.maxerr).max() < 1e-5

    # a case where the refpoint_chk should be 'fail':
    with pytest.warns(la.LinAlgWarning, match=r"Ill\-conditioned matrix"):
        chk2 = cb.cbcoordchk(kaa, b, [25, 26, 27, 31, 32, 33], verbose=False)
    assert chk2.refpoint_chk == "fail"


def compare_nets(net, net2):
    for name in net.__dict__:
        if isinstance(net.__dict__[name], list):
            assert net.__dict__[name] == net2.__dict__[name]
        else:
            assert np.allclose(net.__dict__[name], net2.__dict__[name])


def test_mk_net_drms():
    pth = os.path.dirname(inspect.getfile(cb))
    pth = os.path.join(pth, "tests", "nas2cam_csuper")

    # Load the mass and stiffness from the .op4 file
    # This loads the data into a dict:
    mk = op4.load(os.path.join(pth, "inboard.op4"))
    maa = mk["mxx"][0]
    kaa = mk["kxx"][0]

    # Get the USET table The USET table has the boundary DOF
    # information (id, location, coordinate system). This is needed
    # for superelements with an indeterminate interface. The nastran
    # module has the function bulk2uset which is handy for forming the
    # USET table from bulk data.

    uset, coords = nastran.bulk2uset(os.path.join(pth, "inboard.asm"))

    # uset[::6, [0, 3, 4, 5]]
    # array([[   3.,  600.,    0.,  300.],
    #        [  11.,  600.,  300.,  300.],
    #        [  19.,  600.,  300.,    0.],
    #        [  27.,  600.,    0.,    0.]])

    # x s/c is axial (see figure in cbtf or cbcheck tutorial)
    # - make z l/v axial, but pointing down

    # z s/c  ^ y l/v
    #    \   |   / y s/c
    #       \|/
    #  <------
    # x l/v

    sccoord = [
        [900, 1, 0],
        [0, 0, 0],
        [1, 1, 0],  # z is 45 deg between x & y of l/v
        [0, 0, -1],
    ]  # x is -z l/v
    c = np.cos(45 / 180 * np.pi)
    Tl2s = np.array([[0, 0, -1.0], [-c, c, 0], [c, c, 0]])

    # Form b-set partition vector into a-set
    # In this case, we already know the b-set are first:

    n = uset.shape[0]
    b = np.arange(n)

    # array([ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
    #        16, 17, 18, 19, 20, 21, 22, 23])

    # convert s/c from mm, kg --> m, kg
    ref = [600, 150, 150]
    conv = (0.001, 1.0)
    g = 9.80665

    u = n2p.addgrid(None, 1, "b", sccoord, [0, 0, 0], sccoord)
    Tsc2lv = np.zeros((6, 6))
    T = u.iloc[3:, 1:]
    Tsc2lv[:3, :3] = T
    Tsc2lv[3:, 3:] = T
    assert np.allclose(Tl2s.T, Tsc2lv[:3, :3])

    net = cb.mk_net_drms(
        maa, kaa, b, uset=uset, ref=ref, sccoord=sccoord, conv=conv, g=g
    )

    usetbq, c, bset = nastran.asm2uset(os.path.join(pth, "inboard.asm"))
    with pytest.raises(ValueError, match="number of rows in `uset`"):
        cb.mk_net_drms(
            maa, kaa, b[:3], uset=usetbq, ref=ref, sccoord=Tl2s, conv=conv, g=g
        )

    net2 = cb.mk_net_drms(
        maa, kaa, b, uset=usetbq, ref=ref, sccoord=Tl2s, conv=conv, g=g
    )

    # rb modes in system units:
    uset2, ref2 = cb.uset_convert(uset, ref, conv)
    rb = n2p.rbgeom_uset(uset2, ref2)
    l_sc = net.ifltma_sc[:, :n] @ rb
    l_lv = net.ifltma_lv[:, :n] @ rb
    l_scd = net.ifltmd_sc[:, :n] @ rb
    l_lvd = net.ifltmd_lv[:, :n] @ rb
    a_sc = net.ifatm_sc[:, :n] @ rb
    a_lv = net.ifatm_lv[:, :n] @ rb
    c_sc = net.cgatm_sc[:, :n] @ rb
    c_lv = net.cgatm_lv[:, :n] @ rb

    sbe = np.eye(6)
    sbe[:3] *= 1 / g
    assert np.allclose(a_sc, sbe)

    # calc what the interface forces should be:
    # - acceleration = 1 m/s**2 = 1000 mm/s**2
    # - the forces should be very similar to the 6x6 mass
    #   matrix at the reference point ... which, conveniently,
    #   is provided in the cbtf tutorial:
    mass = np.array(
        [
            [1.755, 0.0, -0.0, 0.0, 0.0, 0.0],
            [0.0, 1.755, -0.0, -0.0, 0.0, 772.22],
            [-0.0, -0.0, 1.755, 0.0, -772.22, -0.0],
            [0.0, -0.0, 0.0, 35905.202, -0.0, -0.0],
            [0.0, 0.0, -772.22, -0.0, 707976.725, 109.558],
            [0.0, 772.22, -0.0, -0.0, 109.558, 707976.725],
        ]
    )
    sbe = mass
    sbe[:, :3] *= 1000  # scale up translations
    assert abs(sbe - l_sc).max() < 0.5

    assert np.allclose(Tsc2lv @ a_sc, a_lv)
    assert np.allclose(Tsc2lv @ c_sc, c_lv)
    assert abs(l_scd).max() < 1e-6 * abs(l_sc).max()
    assert abs(l_lvd).max() < 1e-6 * abs(l_lv).max()
    scale = np.array([[1000], [1000], [1000], [1000000], [1000000], [1000000]])
    assert np.allclose((1 / scale) * (Tsc2lv @ l_sc), l_lv)

    # height and mass values from cbcheck tutorial (and then refined):
    m_kg = 1.75505183
    h_m = 1.039998351 - 0.6
    assert abs(net.height_lv - h_m) < 0.000001
    assert abs(net.weight_lv - m_kg * g) < 0.000001
    assert abs(net.height_sc - 1000 * h_m) < 0.000001 * 1000
    assert abs(net.weight_sc - 1000 * m_kg * g) < 0.000001 * 1000
    assert net.scaxial_sc == 0
    assert net.scaxial_lv == 2

    compare_nets(net, net2)

    # check the natural unit output:
    net3 = cb.mk_net_drms(
        maa, kaa, b, uset=uset, ref=ref, sccoord=Tl2s, conv=conv, g=g, tau=("mm", "m")
    )
    for drm in ("ifatm", "cgatm"):
        if drm == "ifatm":
            # only ifatm has 12 rows
            drm1 = getattr(net, drm)
            drm3 = getattr(net3, drm)
            assert np.allclose(drm3[:3], drm1[:3] * g * 1000)
            assert np.allclose(drm3[3:6], drm1[3:6])
            assert np.allclose(drm3[6:9], drm1[6:9] * g)
            assert np.allclose(drm3[9:], drm1[9:])
        for ext, factor in (("_sc", 1000), ("_lv", 1)):
            drm1 = getattr(net, drm + ext)
            drm3 = getattr(net3, drm + ext)
            assert np.allclose(drm3[:3], drm1[:3] * g * factor)
            assert np.allclose(drm3[3:], drm1[3:])

    labels3 = [i.replace("g", "mm/s^2") for i in net.ifatm_labels[:6]] + [
        i.replace("g", "m/s^2") for i in net.ifatm_labels[6:]
    ]
    print(labels3)
    print()
    print(net3.ifatm_labels)
    assert labels3 == net3.ifatm_labels

    net4 = cb.mk_net_drms(
        maa, kaa, b, uset=uset, ref=ref, sccoord=Tl2s, conv=conv, g=g, tau=("g", "m")
    )
    for drm in ("ifatm", "cgatm"):
        if drm == "ifatm":
            # only ifatm has 12 rows
            drm1 = getattr(net, drm)
            drm3 = getattr(net3, drm)
            drm4 = getattr(net4, drm)
            assert np.allclose(drm4[:3], drm1[:3])
            assert np.allclose(drm4[3:6], drm1[3:6])
            assert np.allclose(drm4[6:9], drm3[6:9])
            assert np.allclose(drm4[9:], drm3[9:])
        for ext, net_ in (("_sc", net), ("_lv", net3)):
            drm1 = getattr(net_, drm + ext)
            drm4 = getattr(net4, drm + ext)
            assert np.allclose(drm4[:3], drm1[:3])
            assert np.allclose(drm4[3:], drm1[3:])

    labels4 = net.ifatm_labels[:6] + net3.ifatm_labels[6:]
    assert labels4 == net4.ifatm_labels

    # test mixed b-set/q-set:
    na = maa.shape[0]
    q = np.r_[n:na]
    newb = np.r_[0:12, na - 12 : na]
    newq = np.r_[12 : na - 12]
    kaa_newb = np.empty((na, na))
    maa_newb = np.empty((na, na))
    for newr, oldr in [(newb, b), (newq, q)]:
        for newc, oldc in [(newb, b), (newq, q)]:
            maa_newb[np.ix_(newr, newc)] = maa[np.ix_(oldr, oldc)]
            kaa_newb[np.ix_(newr, newc)] = kaa[np.ix_(oldr, oldc)]

    net5 = cb.mk_net_drms(
        maa_newb,
        kaa_newb,
        newb,
        uset=uset,
        ref=ref,
        sccoord=sccoord,
        conv=conv,
        g=g,
        reorder=False,
    )

    assert np.allclose(net5.ifltma[:, newb], net.ifltma[:, b])
    assert np.allclose(net5.ifltma[:, newq], net.ifltma[:, q])

    net6 = cb.mk_net_drms(
        maa_newb,
        kaa_newb,
        newb,
        uset=uset,
        ref=ref,
        sccoord=sccoord,
        conv=conv,
        g=g,
        reorder=False,
        rbe3_indep_dof=123456,
    )

    assert np.allclose(net6.ifltma, net5.ifltma)
    # translations match:
    assert np.allclose(net6.ifatm_sc[:3], net5.ifatm_sc[:3])
    # rotations do not:
    assert not np.allclose(net6.ifatm_sc[3:], net5.ifatm_sc[3:])
    vec = ytools.mkpattvec([3, 4, 5], len(newb), 6).ravel()
    assert (net5.ifatm_sc[:, newb[vec]] == 0.0).all()
    assert not (net6.ifatm_sc[:, newb[vec]] == 0.0).all()


def test_mk_net_drms_6dof():
    # same as above, but reduced to single point interface
    pth = os.path.dirname(inspect.getfile(cb))
    pth = os.path.join(pth, "tests", "nas2cam_csuper")
    mk = op4.load(os.path.join(pth, "inboard.op4"))
    maa = mk["mxx"][0]
    kaa = mk["kxx"][0]
    uset, coords = nastran.bulk2uset(os.path.join(pth, "inboard.asm"))
    n = uset.shape[0]
    b = np.arange(n)
    q = np.arange(n, maa.shape[0])
    ttl = maa.shape[0]

    # reduce to single point interface:
    rb = n2p.rbgeom_uset(uset, [600, 150, 150])

    # old = {b_24} = {rb.T @ b_6}  = [rb.T  0_6?] {b_6}
    #       {q_?}    {q_?}           [0_?6   I??] {q_?}
    trans = np.zeros((len(q) + 6, ttl))
    trans[:6, :n] = rb.T
    trans[6:, n:] = np.eye(len(q))
    maa = trans @ maa @ trans.T
    kaa = trans @ kaa @ trans.T

    assert abs(kaa[:6, :6]).max() < 0.02
    kaa[:6, :] = 0.0
    kaa[:, :6] = 0.0

    # no conversion, no coordinate change:
    g = 9.80665
    n = 6
    b = np.arange(n)
    net = cb.mk_net_drms(maa, kaa, b, g=g)
    net2 = cb.mk_net_drms(maa, kaa, b, g=g, bsubset=b)
    l_sc = net.ifltma_sc[:, :n]
    l_lv = net.ifltma_lv[:, :n]
    l_scd = net.ifltmd_sc[:, :n]
    l_lvd = net.ifltmd_lv[:, :n]
    a_sc = net.ifatm_sc[:, :n]
    a_lv = net.ifatm_lv[:, :n]
    c_sc = net.cgatm_sc[:, :n]
    c_lv = net.cgatm_lv[:, :n]

    sbe = np.eye(6)
    sbe[:3] *= 1 / g
    assert np.allclose(a_sc, sbe)
    mass = np.array(
        [
            [1.755, 0.0, -0.0, 0.0, 0.0, 0.0],
            [0.0, 1.755, -0.0, -0.0, 0.0, 772.22],
            [-0.0, -0.0, 1.755, 0.0, -772.22, -0.0],
            [0.0, -0.0, 0.0, 35905.202, -0.0, -0.0],
            [0.0, 0.0, -772.22, -0.0, 707976.725, 109.558],
            [0.0, 772.22, -0.0, -0.0, 109.558, 707976.725],
        ]
    )
    sbe = mass
    assert abs(sbe - l_sc).max() < 0.0005

    Tsc2lv = np.eye(6)
    assert np.allclose(Tsc2lv @ a_sc, a_lv)
    assert np.allclose(Tsc2lv @ c_sc, c_lv)
    assert abs(l_scd).max() < 1e-6 * abs(l_sc).max()
    assert abs(l_lvd).max() < 1e-6 * abs(l_lv).max()
    assert np.allclose((Tsc2lv @ l_sc), l_lv)

    # height and mass values from cbcheck tutorial (and then refined):
    m_kg = 1.75505183
    h_m = 1039.998351 - 600
    assert abs(net.height_lv - h_m) < 0.0001
    assert abs(net.weight_lv - m_kg * g) < 0.0001
    assert abs(net.height_sc - h_m) < 0.0001
    assert abs(net.weight_sc - m_kg * g) < 0.0001
    assert net.scaxial_sc == 0
    assert net.scaxial_lv == 0

    compare_nets(net, net2)


def test_cglf_moment_signs():
    pth = "pyyeti/tests/cla_test_data"

    se = 101
    uset, coords = nastran.bulk2uset(os.path.join(pth, "outboard.asm"))
    dct = op4.read(os.path.join(pth, "outboard.op4"))
    maa = dct["mxx"]
    kaa = dct["kxx"]
    atm = dct["mug1"]
    ltm = dct["mef1"]
    pch = os.path.join(pth, "outboard.pch")
    atm_labels = [
        "Grid {:4d}-{:1d}".format(grid, dof) for grid, dof in nastran.rddtipch(pch)
    ]
    ltm_labels = [
        "CBAR {:4d}-{:1d}".format(cbar, arg)
        for cbar, arg in nastran.rddtipch(pch, "tef1")
    ]

    nb = uset.shape[0]
    nq = maa.shape[0] - nb
    bset = np.arange(nb)
    qset = np.arange(nq) + nb
    ref = [600.0, 150.0, 150.0]
    g = 9806.65

    # use addgrid to get coordinate transformations from lv to sc:
    cid = [1, 0, 0]
    A = [0, 0, 0]
    # define sc in terms of lv coords:
    # (all drawn out by hand)
    BC = [
        [[0, 0, 1.0], [1.0, 0, 0]],  # lv x is up
        [[0, 0, -1.0], [0, 1.0, 0]],  # lv y is up
        [[-1.0, 0, 0.0], [0, 0, 1.0]],  # lv z is up
        [[0, 0, -1.0], [-1.0, 0, 0]],  # lv x is down
        [[0, 0, 1.0], [0, -1.0, 0]],  # lv y is down
        [[0, -1.0, 0], [0, 0, -1.0]],  # lv z is down
    ]
    Ts = []
    nets = []
    rb = n2p.rbgeom_uset(uset, ref)
    rbcglfa = []
    for bc in BC:
        CI = n2p.mkusetcoordinfo([cid, A, *bc], None, {})
        T = CI[2:]
        Ts.append(T)
        net = cb.mk_net_drms(maa, kaa, bset, uset=uset, ref=ref, g=g, sccoord=T)
        nets.append(net)
        rba = net.cglfa[:, :24] @ rb
        rbcglfa.append(rba)
        # sc rows:
        assert np.all(np.sign(rba[1, [1, 5]]) == np.sign(rba[3, [1, 5]]))
        assert np.all(np.sign(rba[2, [2, 4]]) == np.sign(rba[4, [2, 4]]))
        # lv rows:
        assert np.all(np.sign(rba[6, [1, 5]]) == np.sign(rba[8, [1, 5]]))
        assert np.all(np.sign(rba[7, [2, 4]]) == np.sign(rba[9, [2, 4]]))

    wh_sc = nets[0].weight_sc * nets[0].height_sc
    wh_lv = nets[0].weight_lv * nets[0].height_lv
    n = nets[0].cgatm_sc.shape[1]
    # x is down:
    cgdrm = np.vstack(
        (
            # 5 s/c rows
            nets[0].cgatm_sc[:3],
            -nets[0].ifltma_sc[5] / wh_sc,
            nets[0].ifltma_sc[4] / wh_sc,
            # 5 l/v rows
            nets[0].cgatm_lv[:3],
            -nets[0].ifltma_lv[5] / wh_lv,
            nets[0].ifltma_lv[4] / wh_lv,
            # 4 RSS rows ... filled in during data recovery
            np.zeros((4, n)),
        )
    )
    assert np.allclose(cgdrm, nets[0].cglfa)



import numpy as np
from numpy.random import Generator, MT19937
from pyyeti import psd
import scipy.signal as signal
import pytest


# temporary patch for numpy < 2.0
try:
    np.trapezoid
except AttributeError:
    np.trapezoid = np.trapz


def test_get_freq_oct():
    assert np.allclose(
        np.array(psd.get_freq_oct(1, [1, 5])),
        np.array(
            [
                [1.0, 1.99526231, 3.98107171],
                [0.70794578, 1.41253754, 2.81838293],
                [1.41253754, 2.81838293, 5.62341325],
            ]
        ),
    )

    assert np.allclose(
        np.array(psd.get_freq_oct(1, [1, 5], trim="outside")),
        np.array(
            [
                [1.0, 1.99526231, 3.98107171],
                [0.70794578, 1.41253754, 2.81838293],
                [1.41253754, 2.81838293, 5.62341325],
            ]
        ),
    )

    assert np.allclose(
        np.array(psd.get_freq_oct(1, [1, 5], trim="band")),
        np.array(
            [
                [1.0, 1.99526231, 3.98107171],
                [0.70794578, 1.41253754, 2.81838293],
                [1.41253754, 2.81838293, 5.62341325],
            ]
        ),
    )

    assert np.allclose(
        np.array(psd.get_freq_oct(1, [1, 5], trim="inside")),
        np.array([[1.99526231], [1.41253754], [2.81838293]]),
    )

    e = np.array(psd.get_freq_oct(1, [1, 2.5], trim="inside"))
    assert np.all(e.shape == (3, 0))

    assert np.allclose(
        np.array(psd.get_freq_oct(1, [1, 2.5], trim="center")),
        np.array(
            [[1.0, 1.99526231], [0.70794578, 1.41253754], [1.41253754, 2.81838293]]
        ),
    )

    with pytest.raises(ValueError):
        psd.get_freq_oct(1, [1, 2.5], trim="badstring")


def test_proc_psd_spec():
    with pytest.raises(ValueError):
        psd.proc_psd_spec(np.array([1]))
    with pytest.raises(ValueError):
        psd.proc_psd_spec((1, 2, 3))
    with pytest.raises(ValueError):
        psd.proc_psd_spec((np.arange(10), np.random.randn(11)))
    with pytest.raises(ValueError):
        psd.proc_psd_spec((np.arange(10), np.random.randn(10, 2, 2)))


def test_area():
    spec = np.array(
        [
            [20, 0.0053, 0.01],
            [150, 0.04, 0.04],
            [600, 0.04, 0.04],
            [2000, 0.0036, 0.04 * 600 / 2000],
        ]
    )
    # the last value is to trigger the s == -1 logic in area

    spec2 = spec[:, 1:]
    f2 = spec[:, 0]
    areas = psd.area(spec)
    areas2 = psd.area((f2, spec2))

    fi = np.arange(200, 20001) * 0.1
    pi = psd.interp(spec, fi, linear=False)
    areas3 = np.trapezoid(pi, fi, axis=0)

    assert np.allclose(areas, areas2)
    assert np.allclose(areas, areas3)

    s1 = spec[:, :2]
    assert np.all(psd.area(s1) == psd.area((*s1.T,)))


def test_interp():
    spec = np.array([[20, 0.5], [50, 1.0]])
    freq = [15, 35, 60]
    psdlinear = psd.interp(spec, freq, linear=True).ravel()
    psdlog = psd.interp(spec, freq, linear=False).ravel()
    assert np.allclose(psdlinear, [0, 0.75, 0])
    assert np.allclose(psdlog, [0, 0.76352135927358911, 0])


def test_interp_nans():
    Freq = np.arange(1, 10, 2)
    freq = np.arange(1, 10)
    PSD = (Freq**2) / 100

    nan = [np.nan, np.nan]
    Freq = np.hstack((Freq, nan))
    PSD = np.hstack((PSD, nan))

    plog = (freq**2) / 100
    plog2 = psd.interp((Freq, PSD), freq, linear=False)

    assert np.allclose(plog, plog2)


def test_rescale():
    g = np.random.randn(10000)
    sr = 400
    f, p = signal.welch(g, sr, nperseg=sr)
    p3, f3, msv3, ms3 = psd.rescale(p, f)
    p6, f6, msv6, ms6 = psd.rescale(p, f, n_oct=6)
    p6_2, f6_2, msv6_2, ms6_2 = psd.rescale(p, f, freq=f6)
    p12, f12, msv12, ms12 = psd.rescale(p6, f6, n_oct=12)
    i = int(1.0 / 2 ** (1 / 6))
    msv1 = np.sum(p[i:] * (f[1] - f[0]))
    assert abs(msv1 / msv3 - 1) < 0.12
    assert abs(msv1 / msv6 - 1) < 0.06
    assert abs(msv1 / msv12 - 1) < 0.03
    assert np.allclose(p6, p6_2)
    assert np.allclose(f6, f6_2)
    assert np.allclose(msv6, msv6_2)
    assert np.allclose(ms6, ms6_2)
    p_2, f_2, msv_2, ms_2 = psd.rescale(p, f, freq=f)
    assert np.allclose(p, p_2)
    assert np.allclose(f, f_2)
    msv1 = np.sum(p * (f[1] - f[0]))
    assert np.allclose(msv1, msv_2)

    P = np.vstack((p, p)).T
    P3, F3, MSV3, MS3 = psd.rescale(P, f)
    assert np.allclose(p3, P3[:, 0])
    assert np.allclose(f3, F3)
    assert np.allclose(msv3, MSV3[0])
    assert np.allclose(ms3, MS3[:, 0])
    assert np.allclose(p3, P3[:, 1])
    assert np.allclose(msv3, MSV3[1])
    assert np.allclose(ms3, MS3[:, 1])

    in_freq = np.arange(0, 10.1, 0.25)
    out_freq = np.arange(0, 10.1, 5)
    in_p = np.ones_like(in_freq)
    p, f, ms, mvs = psd.rescale(in_p, in_freq, freq=out_freq)
    assert np.allclose(p, [1.0, 1.0, 1.0])
    p, f, ms, mvs = psd.rescale(in_p, in_freq, freq=out_freq, extendends=False)
    assert np.allclose(p, [0.525, 1.0, 0.525])

    p, f, ms, mvs = psd.rescale(
        np.ones(1000), np.arange(1000), freq=np.arange(100, 800, 100), frange=(200, 500)
    )
    assert np.allclose(f, np.arange(200, 600, 100))


def test_spl():
    x = np.random.randn(100000)
    sr = 4000
    f, spl, oaspl = psd.spl(x, sr, sr, timeslice=len(x) / sr)
    # oaspl should be around 170.75 (since variance = 1):
    shouldbe = 10 * np.log10(1 / (2.9e-9) ** 2)
    abs(oaspl / shouldbe - 1) < 0.01
    oaspl1 = oaspl

    f, spl, oaspl = psd.spl(x, sr, sr, fs=0, timeslice=len(x) / sr)
    # oaspl should be around 170.75 (since variance = 1):
    abs(oaspl / shouldbe - 1) < 0.01

    f, spl, oaspl = psd.spl(x, sr)
    # oaspl should be greater than the one above:
    assert oaspl > oaspl1 * 1.01

    with pytest.raises(ValueError):
        psd.spl(x, sr, sr, timeslice=0.5)


def test_psd2time():
    spec = np.array([[20, 0.0768], [50, 0.48], [100, 0.48]])
    sig, sr = psd.psd2time(
        spec, ppc=10, fstart=35, fstop=70, df=0.01, winends=dict(portion=0.01)
    )

    # ensure that expand_method works:
    sig2, sr2 = psd.psd2time(
        spec,
        ppc=10,
        fstart=35,
        fstop=70,
        df=0.01,
        winends=dict(portion=0.01),
        expand_method="rescale",
    )

    f1, p1 = psd.psdmod(sig, sr, timeslice=f"{len(sig)}")
    f2, p2 = psd.psdmod(sig2, sr2, timeslice=f"{len(sig)}")

    assert np.trapezoid(p1, f1) < np.trapezoid(p2, f2)
    with pytest.raises(ValueError):
        psd.psd2time(
            spec,
            ppc=10,
            fstart=35,
            fstop=70,
            df=0.01,
            winends=dict(portion=0.01),
            expand_method="bad expand method",
        )

    multi_spec = np.array([[20, 0.0768, 0.01], [50, 0.48, 0.01], [100, 0.48, 0.01]])
    with pytest.raises(ValueError):
        psd.psd2time(multi_spec, ppc=10, fstart=35, fstop=70, df=0.01)

    assert np.allclose(700.0, sr)  # 70*10
    assert sig.size == 700 * 100
    f, p = signal.welch(sig, sr, nperseg=sr)
    pv = np.logical_and(f >= 37, f <= 68)
    fi = f[pv]
    psdi = p[pv]
    speci = psd.interp(spec, fi).flatten()
    assert abs(speci - psdi).max() < 0.05
    assert abs(np.trapezoid(psdi, fi) - np.trapezoid(speci, fi)) < 0.25

    spec = ([0.1, 5], [0.1, 0.1])
    sig, sr = psd.psd2time(
        spec, ppc=10, fstart=0.1, fstop=5, df=0.2, winends=dict(portion=0.01)
    )
    # df gets internally reset to fstart
    assert np.allclose(5 * 10.0, sr)
    assert sig.size == 50 * 10
    f, p = signal.welch(sig, sr, nperseg=sr)
    pv = np.logical_and(f >= 0.5, f <= 3.0)
    fi = f[pv]
    psdi = p[pv]
    speci = psd.interp(spec, fi).flatten()
    assert abs(speci - psdi).max() < 0.05
    assert abs(np.trapezoid(psdi, fi) - np.trapezoid(speci, fi)) < 0.065

    # ppc gets reset to 2 case:
    spec = np.array([[0.1, 2.0], [5, 2.0]])
    sig, sr = psd.psd2time(
        spec, ppc=1, fstart=0.1, fstop=5, df=0.2, winends=dict(portion=0.01)
    )
    assert (np.sum(np.mean(sig**2)) - 2.0 * 4.9) / (2.0 * 4.9) < 0.1

    # odd length FFT case:
    spec = np.array([[0.1, 2.0], [5, 2.0]])
    sig, sr, t = psd.psd2time(
        spec, ppc=3, fstart=0.2, fstop=5, df=0.2, winends=dict(portion=0.01), gettime=1
    )
    assert np.allclose(5 * 3, sr)
    assert sig.size == 15 * 5
    assert (np.sum(np.mean(sig**2)) - 2.0 * 4.8) / (2.0 * 4.8) < 0.1
    assert np.allclose(t, np.arange(15 * 5) / sr)


def test_psd2time2():
    spec = np.array([[20, 0.0768], [50, 0.48], [100, 0.48]])
    sig1, sr1 = psd.psd2time(
        spec,
        ppc=10,
        fstart=35,
        fstop=70,
        df=0.01,
        winends=dict(portion=0.01),
        rng=Generator(MT19937(1)),
    )

    sig2, sr2 = psd.psd2time(
        spec,
        ppc=10,
        fstart=35,
        fstop=70,
        df=0.01,
        winends=dict(portion=0.01),
        rng=Generator(MT19937(1)),
    )

    assert np.allclose(700.0, sr1)  # 70*10
    assert sig1.size == 700 * 100
    assert (sig1 == sig2).all()
    assert sr2 == sr1


def test_psdmod():
    TF = 30  # make a 30 second signal
    spec = [[20, 50], [1, 1]]
    sig, sr, t = psd.psd2time(
        spec,
        ppc=10,
        fstart=20,
        fstop=50,
        df=1 / TF,
        winends=dict(portion=10),
        gettime=True,
    )
    # sr = 500
    f, p = signal.welch(sig, sr, nperseg=sr)  # 1 second windows, df=1
    f2, p2 = psd.psdmod(sig, sr, nperseg=sr, timeslice=4, tsoverlap=0.5)
    f2b, p2b = psd.psdmod(sig, sr, nperseg=sr, timeslice="2000", tsoverlap=0.5)
    assert np.allclose(f2b, f2)
    assert np.allclose(p2b, p2)

    pv = np.logical_and(f2 > 24, f2 < 47)
    assert np.all(p2[pv] > p[pv])

    # mimic standard welch:
    f3, p3 = psd.psdmod(sig, sr, nperseg=sr, timeslice=30, tsoverlap=0.5)
    assert np.allclose(p3, p)

    # mimic maximax:
    f4, p4 = psd.psdmod(sig, sr, nperseg=sr)
    assert np.all(p4[pv] > p2[pv])

    # test the map output:
    f5, p5, pmap, t = psd.psdmod(sig, sr, getmap=1)
    assert np.allclose(p5, np.max(pmap, axis=1))
    tshouldbe = np.arange(0.5, 30.0 - 0.25, 0.5)
    assert np.allclose(t, tshouldbe)



#!/usr/bin/env python

from argparse import ArgumentParser

from pyyeti.nastran import op4


def main() -> None:
    """Main."""
    parser = ArgumentParser(description="List contents of op4 file(s)")

    parser.add_argument(
        "input_file", metavar="INFILE", help="file to read", type=str, nargs="+"
    )

    args = parser.parse_args()

    for name in args.input_file:
        names, sizes, forms, mtypes = op4.dir(name, verbose=False)
        print()
        print(f"{name}: {len(names)} matrices")
        for n, s, f, m in zip(names, sizes, forms, mtypes):
            print(f"{n:8}, {s[0]:6} x {s[1]:<6}, form={f}, mtype={m}")


if __name__ == "__main__":
    main()



import numpy as np
from setuptools import Extension, setup

# This extension module can not be defined in pyproject.toml because of the required call
# to np.get_include().
setup(
    ext_modules=[
        Extension(
            name="pyyeti.rainflow.c_rain",
            sources=["pyyeti/rainflow/c_rain.c"],
            include_dirs=[np.get_include()],
            optional=True,
        ),
    ],
)
